import os
import posixpath
import numpy as np
from silx.io import get_data
from silx.io.url import DataUrl
from ...utils import copy_dict_items, compare_dicts
from ...io.utils import hdf5_entry_exists, get_h5_value
from ...io.reader import import_h5_to_dict
from ...resources.utils import extract_parameters, get_values_from_file
from ...resources.nxflatfield import update_dataset_info_flats_darks
from ...resources.utils import get_quantities_and_units
from ..estimators import estimate_cor
from ..processconfig import ProcessConfigBase
from .nabu_config import nabu_config, renamed_keys
from .dataset_validator import FullFieldDatasetValidator
from nxtomo.nxobject.nxdetector import ImageKey


class ProcessConfig(ProcessConfigBase):
    """
    A ProcessConfig object has these main fields:
      - dataset_info: information about the current dataset
      - nabu_config: configuration from the user side
      - processing_steps/processing_options: configuration "ready-to use" for underlying classes

    It is built from the following steps.

    (1a) parse config: (conf_fname or conf_dict) --> "nabu_config"

    (1b) browse dataset: (nabu_config or existing dataset_info)  --> dataset_info


    (2) update_dataset_info_with_user_config
      - Update flats/darks
      - CoR  (value or estimation method) # no estimation yet
      - rotation angles
      - translations files
      - user sino normalization (eg. subtraction etc)

    (3) estimations
      - tilt
      - CoR

    (4) coupled validation

    (5) build processing steps

    (6) configure checkpoints (save/resume)
    """

    default_nabu_config = nabu_config
    config_renamed_keys = renamed_keys
    _use_horizontal_translations = True
    _all_processing_steps = [
        "read_chunk",
        "flatfield",
        "ccd_correction",
        "double_flatfield",
        "tilt_correction",
        "phase",
        "unsharp_mask",
        "take_log",
        "radios_movements",  # radios are cropped after this step, if needed
        "sino_normalization",
        "sino_rings_correction",
        "reconstruction",
        "histogram",
        "save",
    ]

    def _update_dataset_info_with_user_config(self):
        """
        Update the 'dataset_info' (DatasetAnalyzer class instance) data structure with options from user configuration.
        """
        self.logger.debug("Updating dataset information with user configuration")
        if self.dataset_info.kind == "nx":
            update_dataset_info_flats_darks(
                self.dataset_info,
                self.nabu_config["preproc"]["flatfield"],
                output_dir=self.nabu_config["output"]["location"],
                darks_flats_dir=self.nabu_config["dataset"]["darks_flats_dir"],
            )
        elif self.dataset_info.kind == "edf":
            self.dataset_info.flats = self.dataset_info.get_reduced_flats()
            self.dataset_info.darks = self.dataset_info.get_reduced_darks()
        self.rec_params = self.nabu_config["reconstruction"]

        subsampling_factor, subsampling_start = self.nabu_config["dataset"]["projections_subsampling"]
        self.subsampling_factor = subsampling_factor or 1
        self.subsampling_start = subsampling_start or 0

        self._update_dataset_with_user_overwrites()
        self._get_rotation_axis_position()
        self._update_rotation_angles()
        self._get_translation_file("reconstruction", "translation_movements_file", "translations")
        self._get_user_sino_normalization()

    def _update_dataset_with_user_overwrites(self):
        user_overwrites = self.nabu_config["dataset"]["overwrite_metadata"].strip()
        if user_overwrites in ("", None):
            return
        possible_overwrites = {"pixel_size": 1e6, "distance": 1.0, "energy": 1.0}
        try:
            overwrites = get_quantities_and_units(user_overwrites)
        except ValueError:
            msg = (
                "Something wrong in config file in 'overwrite_metadata': could not get quantities/units from '%s'. Please check that separators are ';' and that units are provided (separated by a space)"
                % user_overwrites
            )
            self.logger.fatal(msg)
            raise ValueError(msg)
        for quantity, conversion_factor in possible_overwrites.items():
            user_value = overwrites.pop(quantity, None)
            if user_value is not None:
                self.logger.info("Overwriting %s = %s" % (quantity, user_value))
                user_value *= conversion_factor
                setattr(self.dataset_info, quantity, user_value)

    def _get_translation_file(self, config_section, config_key, dataset_info_attr, last_dim=2):
        transl_file = self.nabu_config[config_section][config_key]
        if transl_file in (None, ""):
            return
        translations = None
        if transl_file is not None and "://" not in transl_file:
            try:
                translations = get_values_from_file(
                    transl_file, shape=(self.n_angles(subsampling=False), last_dim), any_size=True
                ).astype(np.float32)
                translations = translations[self.subsampling_start :: self.subsampling_factor]
            except ValueError:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        else:
            try:
                translations = get_data(transl_file)
            except:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        setattr(self.dataset_info, dataset_info_attr, translations)
        if self._use_horizontal_translations and translations is not None:
            # Horizontal translations are handled by "axis_correction" in backprojector
            horizontal_translations = translations[:, 0]
            if np.max(np.abs(horizontal_translations)) > 1e-3:
                self.dataset_info.axis_correction = horizontal_translations

    def _get_rotation_axis_position(self):
        super()._get_rotation_axis_position()
        rec_params = self.nabu_config["reconstruction"]
        axis_correction_file = rec_params["axis_correction_file"]
        axis_correction = None
        if axis_correction_file is not None:
            try:
                axis_correction = get_values_from_file(
                    axis_correction_file,
                    n_values=self.n_angles(subsampling=False),
                    any_size=True,
                ).astype(np.float32)
                axis_correction = axis_correction[self.subsampling_start :: self.subsampling_factor]
            except ValueError:
                print("Something wrong with axis correction file %s" % axis_correction_file)
                raise
        self.dataset_info.axis_correction = axis_correction

    def _update_rotation_angles(self):
        rec_params = self.nabu_config["reconstruction"]
        n_angles = self.dataset_info.n_angles
        angles_file = rec_params["angles_file"]
        if angles_file is not None:
            try:
                angles = get_values_from_file(angles_file, n_values=n_angles, any_size=True)
                angles = np.deg2rad(angles)
            except ValueError:
                self.logger.fatal("Something wrong with angle file %s" % angles_file)
                raise
            self.dataset_info.rotation_angles = angles
        elif self.dataset_info.rotation_angles is None:
            angles_range_txt = "[0, 180[ degrees"
            if rec_params["enable_halftomo"]:
                angles_range_txt = "[0, 360] degrees"
                angles = np.linspace(0, 2 * np.pi, n_angles, True)
            else:
                angles = np.linspace(0, np.pi, n_angles, False)
            self.logger.warning(
                "No information was found on rotation angles. Using default %s (half tomo is %s)"
                % (angles_range_txt, {0: "OFF", 1: "ON"}[int(self.do_halftomo)])
            )
            self.dataset_info.rotation_angles = angles

    def _get_cor(self):
        cor = self.nabu_config["reconstruction"]["rotation_axis_position"]
        if isinstance(cor, str):  # auto-CoR
            cor = estimate_cor(
                cor,
                self.dataset_info,
                do_flatfield=(self.nabu_config["preproc"]["flatfield"] is not False),
                cor_options=self.nabu_config["reconstruction"]["cor_options"],
                logger=self.logger,
            )
            self.logger.info("Estimated center of rotation: %.3f" % cor)
        self.dataset_info.axis_position = cor

    def _get_user_sino_normalization(self):
        self._sino_normalization_arr = None
        norm = self.nabu_config["preproc"]["sino_normalization"]
        if norm not in ["subtraction", "division"]:
            return
        norm_path = "silx://" + self.nabu_config["preproc"]["sino_normalization_file"].strip()
        url = DataUrl(norm_path)
        try:
            norm_array = get_data(url)
            self._sino_normalization_arr = norm_array.astype("f")
        except (ValueError, OSError) as exc:
            error_msg = "Could not load sino_normalization_file %s. The error was: %s" % (norm_path, str(exc))
            self.logger.error(error_msg)
            raise ValueError(error_msg)

    @property
    def do_halftomo(self):
        """
        Return True if the current dataset is to be reconstructed using 'half-acquisition' setting.
        """
        enable_halftomo = self.nabu_config["reconstruction"]["enable_halftomo"]
        is_halftomo_dataset = self.dataset_info.is_halftomo
        if enable_halftomo == "auto":
            if is_halftomo_dataset is None:
                raise ValueError(
                    "enable_halftomo was set to 'auto', but information on field of view was not found. Please set either 0 or 1 for enable_halftomo"
                )
            return is_halftomo_dataset
        return enable_halftomo

    def _coupled_validation(self):
        self.logger.debug("Doing coupled validation")
        self._dataset_validator = FullFieldDatasetValidator(self.nabu_config, self.dataset_info)
        # Not so ideal to propagate fields like this
        for what in ["rec_params", "rec_region", "binning"]:
            setattr(self, what, getattr(self._dataset_validator, what))

    #
    # Attributes that combine dataset information and user overwrites (eg. binning).
    # Must be accessed after __init__() is done
    #

    @property
    def binning_x(self):
        return getattr(self, "binning", (1, 1))[0]

    @property
    def binning_z(self):
        return getattr(self, "binning", (1, 1))[1]

    @property
    def subsampling(self):
        return getattr(self, "subsampling_factor", None)

    def radio_shape(self, binning=False):
        n_x, n_z = self.dataset_info.radio_dims
        if binning:
            n_z //= self.binning_z
            n_x //= self.binning_x
        return (n_z, n_x)

    def n_angles(self, subsampling=False):
        rot_angles = self.dataset_info.rotation_angles
        if subsampling:
            rot_angles = rot_angles[self.subsampling_start :: self.subsampling_factor]
        return len(rot_angles)

    def radios_shape(self, binning=False, subsampling=False):
        n_z, n_x = self.radio_shape(binning=binning)
        n_a = self.n_angles(subsampling=subsampling)
        return (n_a, n_z, n_x)

    def rotation_axis_position(self, binning=False):
        cor = self.dataset_info.axis_position  # might be None (default to the middle of detector)
        if cor is None and self.do_halftomo:
            raise ValueError("No information on center of rotation, cannot use half-tomography reconstruction")
        if cor is not None and binning:
            # Backprojector uses middle of pixel for coordinate indices.
            # This means that the leftmost edge of the leftmost pixel has coordinate -0.5.
            # When using binning with a factor 'b', the CoR has to adapted as
            #   cor_binned  =  (cor + 0.5)/b - 0.5
            cor = (cor + 0.5) / self.binning_x - 0.5
        return cor

    def sino_shape(self, binning=False, subsampling=False):
        """
        Return the shape of a sinogram image.

        Parameter
        ---------
        binning: bool
            Whether to account for image binning
        subsampling: bool
            Whether to account for projections subsampling
        """
        n_a, _, n_x = self.radios_shape(binning=binning, subsampling=subsampling)
        return (n_a, n_x)

    def sinos_shape(self, binning=False, subsampling=False):
        n_z, _ = self.radio_shape(binning=binning)
        return (n_z,) + self.sino_shape(binning=binning, subsampling=subsampling)

    def projs_indices(self, subsampling=False):
        step = 1
        if subsampling:
            step = self.subsampling or 1
        return sorted(self.dataset_info.projections.keys())[::step]

    def rotation_angles(self, subsampling=False):
        start = 0
        step = 1
        if subsampling:
            start = self.subsampling_start
            step = self.subsampling_factor
        return self.dataset_info.rotation_angles[start::step]

    @property
    def rec_roi(self):
        """
        Returns the reconstruction region of interest (ROI), accounting for binning in both dimensions.
        """
        rec_params = self.rec_params  # accounts for binning
        x_s, x_e = rec_params["start_x"], rec_params["end_x"]
        y_s, y_e = rec_params["start_y"], rec_params["end_y"]
        # Upper bound (end_xy) is INCLUDED in nabu config, hence the +1 here
        return (x_s, x_e + 1, y_s, y_e + 1)

    @property
    def rec_shape(self):
        # Accounts for binning!
        return tuple(np.diff(self.rec_roi)[::-2])

    @property
    def rec_delta_z(self):
        # Accounts for binning!
        z_s, z_e = self.rec_params["start_z"], self.rec_params["end_z"]
        # Upper bound (end_xy) is INCLUDED in nabu config, hence the +1 here
        return z_e + 1 - z_s

    def is_before_radios_cropping(self, step):
        """
        Return true if a given processing step happens before radios cropping
        """
        if step == "sinogram":
            return False
        if step not in self._all_processing_steps:
            raise ValueError("Unknown step: '%s'. Available are: %s" % (step, self._all_processing_steps))
        # sino_normalization
        return self._all_processing_steps.index(step) <= self._all_processing_steps.index("radios_movements")

    #
    # Build processing steps
    #

    # TODO update behavior and remove this function once GroupedPipeline cuda backend is implemented
    def get_radios_rotation_mode(self):
        """
        Determine whether projections are to be rotated, and if so, when they are to be rotated.

        Returns
        -------
        method: str or None
            Rotation method: one of the values of `nabu.resources.params.radios_rotation_mode`
        """
        tilt = self.dataset_info.detector_tilt
        phase_method = self.nabu_config["phase"]["method"]
        do_ctf = phase_method == "CTF"
        do_pag = phase_method == "paganin"
        do_unsharp = (
            self.nabu_config["phase"]["unsharp_method"] is not None and self.nabu_config["phase"]["unsharp_coeff"] > 0
        )
        if tilt is None:
            return None
        if do_ctf:
            return "full"
        # TODO "chunked" rotation is done only when using a "processing margin"
        # For now the processing margin is enabled only if phase or unsharp is enabled.
        # We can either
        #   - Enable processing margin if rotating projections is needed (more complicated to implement)
        #   - Always do "full" rotation (simpler to implement, at the expense of performances)
        if do_pag or do_unsharp:
            return "chunk"
        else:
            return "full"

    def _build_processing_steps(self):
        nabu_config = self.nabu_config
        dataset_info = self.dataset_info
        binning = (nabu_config["dataset"]["binning"], nabu_config["dataset"]["binning_z"])
        tasks = []
        options = {}

        #
        # Dataset / Get data
        #
        # First thing to do is to get the data (radios or sinograms)
        # For now data is assumed to be on disk (see issue #66).
        tasks.append("read_chunk")
        options["read_chunk"] = {
            "sub_region": None,
            "binning": binning,
            "dataset_subsampling": nabu_config["dataset"]["projections_subsampling"],
        }
        #
        # Flat-field
        #
        if nabu_config["preproc"]["flatfield"]:
            tasks.append("flatfield")
            options["flatfield"] = {
                #  Data reader handles binning/subsampling by itself,
                # but FlatField needs "real" indices (after binning/subsampling)
                "projs_indices": self.projs_indices(subsampling=False),
                "binning": binning,
                "do_flat_distortion": nabu_config["preproc"]["flat_distortion_correction_enabled"],
                "flat_distortion_params": extract_parameters(nabu_config["preproc"]["flat_distortion_params"]),
            }
            normalize_srcurrent = nabu_config["preproc"]["normalize_srcurrent"]
            radios_srcurrent = None
            flats_srcurrent = None
            if normalize_srcurrent:
                if (
                    dataset_info.projections_srcurrent is None
                    or dataset_info.flats_srcurrent is None
                    or len(dataset_info.flats_srcurrent) == 0
                ):
                    self.logger.error("Cannot do SRCurrent normalization: missing flats and/or projections SRCurrent")
                    normalize_srcurrent = False
                else:
                    radios_srcurrent = dataset_info.projections_srcurrent
                    flats_srcurrent = dataset_info.flats_srcurrent
            options["flatfield"].update(
                {
                    "normalize_srcurrent": normalize_srcurrent,
                    "radios_srcurrent": radios_srcurrent,
                    "flats_srcurrent": flats_srcurrent,
                }
            )
            if len(dataset_info.darks) > 1:
                self.logger.warning("Cannot do flat-field with more than one reduced dark. Taking the first one.")
                dataset_info.darks = dataset_info.darks[sorted(dataset_info.darks.keys())[0]]
        #
        # Spikes filter
        #
        if nabu_config["preproc"]["ccd_filter_enabled"]:
            tasks.append("ccd_correction")
            options["ccd_correction"] = {
                "type": "median_clip",  # only one available for now
                "median_clip_thresh": nabu_config["preproc"]["ccd_filter_threshold"],
            }
        #
        # Double flat field
        #
        if nabu_config["preproc"]["double_flatfield_enabled"]:
            tasks.append("double_flatfield")
            options["double_flatfield"] = {
                "sigma": nabu_config["preproc"]["dff_sigma"],
                "processes_file": nabu_config["preproc"]["processes_file"],
                "log_min_clip": nabu_config["preproc"]["log_min_clip"],
                "log_max_clip": nabu_config["preproc"]["log_max_clip"],
            }
        #
        # Radios rotation (do it here if possible)
        #
        if self.get_radios_rotation_mode() == "chunk":
            tasks.append("tilt_correction")
            options["tilt_correction"] = {
                "angle": dataset_info.detector_tilt,
                "center": nabu_config["preproc"]["rotate_projections_center"],
                "mode": "chunk",
            }
        #
        #
        # Phase retrieval
        #
        if nabu_config["phase"]["method"] is not None:
            tasks.append("phase")
            options["phase"] = copy_dict_items(nabu_config["phase"], ["method", "delta_beta", "padding_type"])
            options["phase"].update(
                {
                    "energy_kev": dataset_info.energy,
                    "distance_cm": dataset_info.distance * 1e2,
                    "distance_m": dataset_info.distance,
                    "pixel_size_microns": dataset_info.pixel_size,
                    "pixel_size_m": dataset_info.pixel_size * 1e-6,
                }
            )
            if binning != (1, 1):
                options["phase"]["delta_beta"] /= binning[0] * binning[1]
            if options["phase"]["method"] == "CTF":
                self._get_ctf_parameters(options["phase"])
        #
        # Unsharp
        #
        if (
            nabu_config["phase"]["unsharp_method"] is not None
            and nabu_config["phase"]["unsharp_coeff"] > 0
            and nabu_config["phase"]["unsharp_sigma"] > 0
        ):
            tasks.append("unsharp_mask")
            options["unsharp_mask"] = copy_dict_items(
                nabu_config["phase"], ["unsharp_coeff", "unsharp_sigma", "unsharp_method"]
            )
        #
        # -logarithm
        #
        if nabu_config["preproc"]["take_logarithm"]:
            tasks.append("take_log")
            options["take_log"] = copy_dict_items(nabu_config["preproc"], ["log_min_clip", "log_max_clip"])
        #
        # Radios rotation (do it here if mode=="full")
        #
        if self.get_radios_rotation_mode() == "full":
            tasks.append("tilt_correction")
            options["tilt_correction"] = {
                "angle": dataset_info.detector_tilt,
                "center": nabu_config["preproc"]["rotate_projections_center"],
                "mode": "full",
            }
        #
        # Translation movements
        #

        translations = dataset_info.translations
        if translations is not None:
            if np.max(np.abs(translations[:, 1])) < 1e-5:
                self.logger.warning("No vertical translation greater than 1e-5 - disabling vertical shifts")
                # horizontal movements are handled in backprojector
            else:
                tasks.append("radios_movements")
                options["radios_movements"] = {"translation_movements": dataset_info.translations}
        #
        # Sinogram normalization (before half-tomo)
        #
        if nabu_config["preproc"]["sino_normalization"] is not None:
            tasks.append("sino_normalization")
            options["sino_normalization"] = {
                "method": nabu_config["preproc"]["sino_normalization"],
                "normalization_array": self._sino_normalization_arr,
            }

        #
        # Sinogram-based rings artefacts removal
        #
        if nabu_config["preproc"]["sino_rings_correction"]:
            tasks.append("sino_rings_correction")
            options["sino_rings_correction"] = {
                "method": nabu_config["preproc"]["sino_rings_correction"],
                "user_options": nabu_config["preproc"]["sino_rings_options"],
            }
        #
        # Reconstruction
        #
        if nabu_config["reconstruction"]["method"] is not None:
            tasks.append("reconstruction")
            # Iterative is not supported through configuration file for now.
            options["reconstruction"] = copy_dict_items(
                self.rec_params,
                [
                    "method",
                    "implementation",
                    "fbp_filter_type",
                    "fbp_filter_cutoff",
                    "padding_type",
                    "start_x",
                    "end_x",
                    "start_y",
                    "end_y",
                    "start_z",
                    "end_z",
                    "centered_axis",
                    "clip_outer_circle",
                    "outer_circle_value",
                    "source_sample_dist",
                    "sample_detector_dist",
                    "hbp_legs",
                    "hbp_reduction_steps",
                ],
            )
            rec_options = options["reconstruction"]
            rec_options["rotation_axis_position"] = self.rotation_axis_position(binning=True)
            rec_options["enable_halftomo"] = self.do_halftomo
            rec_options["axis_correction"] = dataset_info.axis_correction
            if dataset_info.axis_correction is not None:
                rec_options["axis_correction"] = rec_options["axis_correction"]

            rec_options["angles"] = np.array(self.rotation_angles(subsampling=True))
            rec_options["angles"] += np.deg2rad(nabu_config["reconstruction"]["angle_offset"])
            voxel_size = dataset_info.pixel_size * 1e-4
            rec_options["voxel_size_cm"] = (
                voxel_size,
                voxel_size,
                voxel_size,
            )  # pix size is in microns in dataset_info

            rec_options["iterations"] = nabu_config["reconstruction"]["iterations"]

            # x/y/z position information
            def get_mean_pos(position_array):
                if position_array is None:
                    return None
                else:
                    position_array = np.array(position_array)
                    # filter only projections. Avoid getting noise
                    position_array = position_array[
                        np.asarray(dataset_info.dataset_scanner.image_key_control) == ImageKey.PROJECTION.value
                    ]
                    return float(np.mean(position_array))

            mean_positions_xyz = (
                get_mean_pos(dataset_info.dataset_scanner.z_translation),
                get_mean_pos(dataset_info.dataset_scanner.y_translation),
                get_mean_pos(dataset_info.dataset_scanner.x_translation),
            )
            if all([m is not None for m in mean_positions_xyz]):
                rec_options["position"] = mean_positions_xyz
            if rec_options["method"] == "cone" and rec_options["sample_detector_dist"] is None:
                rec_options["sample_detector_dist"] = self.dataset_info.distance  # was checked to be not None earlier

            # New key
            rec_options["cor_estimated_auto"] = isinstance(nabu_config["reconstruction"]["rotation_axis_position"], str)
        #
        # Histogram
        #
        if nabu_config["postproc"]["output_histogram"]:
            tasks.append("histogram")
            options["histogram"] = copy_dict_items(nabu_config["postproc"], ["histogram_bins"])
        #
        # Save
        #
        if nabu_config["output"]["location"] is not None:
            tasks.append("save")
            options["save"] = copy_dict_items(nabu_config["output"], list(nabu_config["output"].keys()))
            options["save"]["overwrite"] = nabu_config["output"]["overwrite_results"]

        self.processing_steps = tasks
        self.processing_options = options
        if set(self.processing_steps) != set(self.processing_options.keys()):
            raise ValueError("Something wrong with process_config: options do not correspond to steps")
        # Add check
        if set(self.processing_steps) != set(self.processing_options.keys()):
            raise ValueError("Something wrong when building processing steps")

    def _get_ctf_parameters(self, phase_options):
        dataset_info = self.dataset_info
        user_phase_options = self.nabu_config["phase"]

        ctf_geom = extract_parameters(user_phase_options["ctf_geometry"])
        ctf_advanced_params = extract_parameters(user_phase_options["ctf_advanced_params"])

        # z1_vh
        z1_v = ctf_geom["z1_v"]
        z1_h = ctf_geom["z1_h"]
        z1_vh = None
        if z1_h is None and z1_v is None:
            # parallel beam
            z1_vh = None
        elif (z1_v is None) ^ (z1_h is None):
            # only one is provided: source-sample distance
            z1_vh = z1_v or z1_h
        if z1_h is not None and z1_v is not None:
            # distance of the vertically focused source (horizontal line) and the horizontaly focused source (vertical line)
            # for KB mirrors
            z1_vh = (z1_v, z1_h)
        # pix_size_det
        pix_size_det = ctf_geom["detec_pixel_size"] or dataset_info.pixel_size * 1e-6
        # wavelength
        wavelength = 1.23984199e-9 / dataset_info.energy

        phase_options["ctf_geo_pars"] = {
            "z1_vh": z1_vh,
            "z2": phase_options["distance_m"],
            "pix_size_det": pix_size_det,
            "wavelength": wavelength,
            "magnification": bool(ctf_geom["magnification"]),
            "length_scale": ctf_advanced_params["length_scale"],
        }
        phase_options["ctf_lim1"] = ctf_advanced_params["lim1"]
        phase_options["ctf_lim2"] = ctf_advanced_params["lim2"]
        phase_options["ctf_normalize_by_mean"] = ctf_advanced_params["normalize_by_mean"]

    def _configure_save_steps(self, steps_to_save=None):
        self.steps_to_save = []
        self.dump_sinogram = False
        if steps_to_save is None:
            steps_to_save = self.nabu_config["pipeline"]["save_steps"]
            if steps_to_save in (None, ""):
                return
            steps_to_save = [s.strip() for s in steps_to_save.split(",")]
        for step in self.processing_steps:
            step = step.strip()
            if step in steps_to_save:
                self.processing_options[step]["save"] = True
                self.processing_options[step]["save_steps_file"] = self.get_save_steps_file(step_name=step)
        # "sinogram" is a special keyword, not explicitly in the processing steps
        if "sinogram" in steps_to_save:
            self.dump_sinogram = True
            self.dump_sinogram_file = self.get_save_steps_file(step_name="sinogram")
        self.steps_to_save = steps_to_save

    def _get_dump_file_and_h5_path(self):
        resume_from = self.resume_from_step
        process_file = self.get_save_steps_file(step_name=resume_from)
        if not os.path.isfile(process_file):
            self.logger.error("Cannot resume processing from step '%s': no such file %s" % (resume_from, process_file))
            return None, None
        h5_entry = self.dataset_info.hdf5_entry or "entry"
        process_h5_path = posixpath.join(h5_entry, resume_from, "results/data")
        if not hdf5_entry_exists(process_file, process_h5_path):
            self.logger.error("Could not find data in %s in file %s" % (process_h5_path, process_file))
            process_h5_path = None
        return process_file, process_h5_path

    def _configure_resume(self, resume_from=None):
        self.resume_from_step = None
        if resume_from is None:
            resume_from = self.nabu_config["pipeline"]["resume_from_step"]
            if resume_from in (None, ""):
                return
            resume_from = resume_from.strip(" ,;")
        self.resume_from_step = resume_from

        processing_steps = self.processing_steps
        # special case: resume from sinogram
        if resume_from == "sinogram":
            # disable up to 'reconstruction', not included
            if "sino_rings_correction" in processing_steps:
                # Sinogram destriping is done before building the half tomo sino.
                # Not sure if this is needed (i.e can we do before building the extended sino ?)
                # TODO find a more elegant way
                idx = processing_steps.index("sino_rings_correction")
            else:
                idx = processing_steps.index("reconstruction")
        #
        elif resume_from in processing_steps:
            idx = processing_steps.index(resume_from) + 1  # disable up to resume_from, included
        else:
            msg = "Cannot resume processing from step '%s': no such step in the current configuration" % resume_from
            self.logger.error(msg)
            self.resume_from_step = None
            return

        # Get corresponding file and h5 path
        process_file, process_h5_path = self._get_dump_file_and_h5_path()
        if process_file is None or process_h5_path is None:
            self.resume_from_step = None
            return
        dump_info = self._check_dump_file(process_file, raise_on_error=False)
        if dump_info is None:
            self.logger.error("Cannot resume from step %s: cannot use file %s" % (resume_from, process_file))
            self.resume_from_step = None
            return
        dump_start_z, dump_end_z = dump_info

        # Disable steps
        steps_to_disable = processing_steps[1:idx]
        self.logger.debug("Disabling steps %s" % str(steps_to_disable))
        for step_name in steps_to_disable:
            processing_steps.remove(step_name)
            self.processing_options.pop(step_name)

        # Update configuration
        self.logger.info("Processing will be resumed from step '%s' using file %s" % (resume_from, process_file))
        self._old_read_chunk = self.processing_options["read_chunk"]
        self.processing_options["read_chunk"] = {
            "process_file": process_file,
            "process_h5_path": process_h5_path,
            "step_name": resume_from,
            "dump_start_z": dump_start_z,
            "dump_end_z": dump_end_z,
        }
        # Dont dump a step if we resume from this step
        if resume_from in self.steps_to_save:
            self.logger.warning(
                "Processing is resumed from step '%s'. This step won't be dumped to a file" % resume_from
            )
            self.steps_to_save.remove(resume_from)
            if resume_from == "sinogram":
                self.dump_sinogram = False
            else:
                if resume_from in self.processing_options:  # should not happen
                    self.processing_options[resume_from].pop("save")

    def _check_dump_file(self, process_file, raise_on_error=False):
        """
        Return (start_z, end_z) on success
        Return None on failure
        """
        # Ensure data in the file correspond to what is currently asked
        if self.resume_from_step is None:
            return None

        # Check dataset shape/start_z/end_z
        rec_cfg_h5_path = posixpath.join(
            self.dataset_info.hdf5_entry or "entry",
            self.resume_from_step,
            "configuration/processing_options/reconstruction",
        )
        dump_start_z = get_h5_value(process_file, posixpath.join(rec_cfg_h5_path, "start_z"))
        dump_end_z = get_h5_value(process_file, posixpath.join(rec_cfg_h5_path, "end_z"))
        if dump_end_z < 0:
            dump_end_z += self.radio_shape(binning=False)[0]
        start_z, end_z = (
            self.processing_options["reconstruction"]["start_z"],
            self.processing_options["reconstruction"]["end_z"],
        )
        if not (dump_start_z <= start_z and end_z <= dump_end_z):
            msg = (
                "File %s was built with start_z=%d, end_z=%d but current configuration asks for start_z=%d, end_z=%d"
                % (process_file, dump_start_z, dump_end_z, start_z, end_z)
            )
            if not raise_on_error:
                self.logger.error(msg)
                return None
            self.logger.fatal(msg)
            raise ValueError(msg)

        # Check parameters other than reconstruction
        filedump_nabu_config = import_h5_to_dict(
            process_file,
            posixpath.join(self.dataset_info.hdf5_entry or "entry", self.resume_from_step, "configuration/nabu_config"),
        )
        sections_to_ignore = ["reconstruction", "output", "pipeline"]
        for section in sections_to_ignore:
            filedump_nabu_config[section] = self.nabu_config[section]
        # special case of the "save_steps process"
        # filedump_nabu_config["pipeline"]["save_steps"] = self.nabu_config["pipeline"]["save_steps"]

        diff = compare_dicts(filedump_nabu_config, self.nabu_config)
        if diff is not None:
            msg = "Nabu configuration in file %s differ from the current one: %s" % (process_file, diff)
            if not raise_on_error:
                self.logger.error(msg)
                return None
            self.logger.fatal(msg)
            raise ValueError(msg)
        #

        return (dump_start_z, dump_end_z)

    def get_save_steps_file(self, step_name=None):
        if self.nabu_config["pipeline"]["steps_file"] not in (None, ""):
            return self.nabu_config["pipeline"]["steps_file"]
        nabu_save_options = self.nabu_config["output"]
        output_dir = nabu_save_options["location"]
        file_prefix = step_name + "_" + nabu_save_options["file_prefix"]
        fname = os.path.join(output_dir, file_prefix) + ".hdf5"
        return fname



import os
import copy
from typing import Optional, Union
import numpy
from silx.io.url import DataUrl
from tomoscan.tomoobject import TomoObject
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf import EDFTomoScan
from tomoscan.esrf.volume import HDF5Volume, MultiTIFFVolume
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from ..app.bootstrap_stitching import _SECTIONS_COMMENTS
from ..pipeline.config import generate_nabu_configfile
from .config import (
    StitchingConfiguration,
    get_default_stitching_config,
    PreProcessedSingleAxisStitchingConfiguration,
    PostProcessedSingleAxisStitchingConfiguration,
    SLURM_SECTION,
)

try:
    from sluurp.job import SBatchScriptJob
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True


def split_stitching_configuration_to_slurm_job(
    configuration: StitchingConfiguration, yield_configuration: bool = False
):
    """
    generator to split a StitchingConfiguration into several SBatchScriptJob.

    This will handle:
       * division into several jobs according to `slices` and `n_job`
       * creation of SBatchScriptJob handling slurm configuration and command to be launched

    :param StitchingConfiguration configuration: configuration of the stitching to launch (into several jobs)
    :param bool yield_configuration: if True then yield (SBatchScriptJob, StitchingConfiguration) else yield only SBatchScriptJob
    """
    if not isinstance(configuration, StitchingConfiguration):
        raise TypeError(
            f"configuration is expected to be an instance of {StitchingConfiguration}. {type(configuration)} provided."
        )
    if not has_sluurp:
        raise ImportError("sluurp not install. Please install it to distribute stitching on slurm (pip install sluurm)")

    slurm_configuration = configuration.slurm_config
    n_jobs = slurm_configuration.n_jobs
    stitching_type = configuration.stitching_type
    # cleqn slurm configurqtion
    slurm_configuration = slurm_configuration.to_dict()
    clean_script = slurm_configuration.pop("clean_scripts", False)
    slurm_configuration.pop("n_jobs", None)
    # for now other_options is not handled
    slurm_configuration.pop("other_options", None)

    if "memory" in slurm_configuration and isinstance(slurm_configuration["memory"], str):
        memory = slurm_configuration["memory"].lower().replace(" ", "")
        memory = memory.rstrip("b").rstrip("g")
        slurm_configuration["memory"] = memory
    # handle slices if None
    configuration.settle_inputs()

    slice_sub_parts = split_slices(slices=configuration.slices, n_parts=n_jobs)
    if isinstance(configuration, PreProcessedSingleAxisStitchingConfiguration):
        stitch_prefix = os.path.basename(os.path.splitext(configuration.output_file_path)[0])
        configuration.output_file_path = os.path.abspath(configuration.output_file_path)

    elif isinstance(configuration, PostProcessedSingleAxisStitchingConfiguration):
        stitch_prefix = os.path.basename(os.path.splitext(configuration.output_volume.file_path)[0])
        configuration.output_volume.file_path = os.path.abspath(configuration.output_volume.file_path)
    else:
        raise TypeError(f"{type(configuration)} not handled")

    for i_sub_part, slice_sub_part in enumerate(slice_sub_parts):
        sub_configuration = copy.deepcopy(configuration)
        # update slice
        sub_configuration.slices = slice_sub_part
        # remove slurm configuration because once on the partition we run it manually
        sub_configuration.slurm_config = None

        if isinstance(sub_configuration, PreProcessedSingleAxisStitchingConfiguration):
            original_output_file_path, file_extension = os.path.splitext(sub_configuration.output_file_path)
            sub_configuration.output_file_path = os.path.join(
                original_output_file_path,
                os.path.basename(original_output_file_path) + f"_part_{i_sub_part}" + file_extension,
            )
            output_obj = NXtomoScan(
                scan=sub_configuration.output_file_path,
                entry=sub_configuration.output_data_path,
            )
        elif isinstance(sub_configuration, PostProcessedSingleAxisStitchingConfiguration):
            if isinstance(sub_configuration.output_volume, (HDF5Volume, MultiTIFFVolume)):
                original_output_file_path, file_extension = os.path.splitext(sub_configuration.output_volume.file_path)
                sub_configuration.output_volume.file_path = os.path.join(
                    original_output_file_path,
                    os.path.basename(original_output_file_path) + f"_part_{i_sub_part}" + file_extension,
                )
            elif isinstance(sub_configuration.output_volume, VolumeSingleFrameBase):
                url = sub_configuration.output_volume.url
                original_output_folder = url.file_path()
                sub_part_url = DataUrl(
                    file_path=os.path.join(
                        original_output_folder,
                        os.path.basename(original_output_folder) + f"_part_{i_sub_part}",
                    ),
                    data_path=url.data_path(),
                    scheme=url.scheme(),
                    data_slice=url.data_slice(),
                )
                sub_configuration.output_volume.url = sub_part_url
            output_obj = sub_configuration.output_volume
        else:
            raise TypeError(f"{type(sub_configuration)} not handled")

        working_directory = get_working_directory(output_obj)
        if working_directory is not None:
            script_dir = working_directory
        else:
            script_dir = "./"

        # save sub part nabu configuration file
        slurm_script_name = f"{stitch_prefix}_part_{i_sub_part}.sh"
        nabu_script_name = f"{stitch_prefix}_part_{i_sub_part}.conf"

        os.makedirs(script_dir, exist_ok=True)
        default_config = get_default_stitching_config(stitching_type)
        default_config.pop(SLURM_SECTION, None)
        generate_nabu_configfile(
            fname=os.path.join(script_dir, nabu_script_name),
            default_config=default_config,
            comments=True,
            sections_comments=_SECTIONS_COMMENTS,
            prefilled_values=sub_configuration.to_dict(),
            options_level="advanced",
        )

        command = f"python3 -m nabu.app.stitching {os.path.join(script_dir, nabu_script_name)}"
        script = (command,)

        job = SBatchScriptJob(
            slurm_config=slurm_configuration,
            script=script,
            script_path=os.path.join(script_dir, slurm_script_name),
            clean_script=clean_script,
            working_directory=working_directory,
        )
        job.overwrite = True
        if yield_configuration:
            yield job, sub_configuration
        else:
            yield job


def split_slices(slices: Union[slice, tuple], n_parts: int):
    if not isinstance(n_parts, int):
        raise TypeError(f"n_parts should be an int. {type(n_parts)} provided")
    if isinstance(slices, slice):
        assert isinstance(slices.start, int), "slices.start must be an integer"
        assert isinstance(slices.stop, int), "slices.stop must be an integer"
        start = stop = slices.start
        steps_size = int(numpy.ceil((slices.stop - slices.start) / n_parts))
        while stop < slices.stop:
            stop = min(start + steps_size, slices.stop)
            yield slice(start, stop, slices.step)
            start = stop
    elif isinstance(slices, (tuple, list)):
        start = stop = 0
        steps_size = int(numpy.ceil(len(slices) / n_parts))
        while stop < len(slices):
            stop = min(start + steps_size, len(slices))
            yield slices[start:stop]
            start = stop
    else:
        raise TypeError(f"slices type ({type(slices)}) is not handled. Must be a slice or an Iterable")


def get_working_directory(obj: TomoObject) -> Optional[str]:
    """
    return working directory for a specific TomoObject
    """
    if not isinstance(obj, TomoObject):
        raise TypeError(f"obj should be an instance of {TomoObject}. {type(obj)} provided")
    if isinstance(obj, (HDF5Volume, MultiTIFFVolume)):
        if obj.file_path is None:
            return None
        else:
            return os.path.abspath(os.path.dirname(obj.file_path))
    elif isinstance(obj, VolumeSingleFrameBase):
        if obj.data_url is not None:
            return os.path.abspath(obj.data_url.file_path())
        else:
            return None
    elif isinstance(obj, EDFTomoScan):
        return obj.path
    elif isinstance(obj, NXtomoScan):
        if obj.master_file is None:
            return None
        else:
            return os.path.abspath(os.path.dirname(obj.master_file))
    else:
        raise RuntimeError(f"obj type not handled ({type(obj)})")



from ..resources.utils import extract_parameters
from ..io.detector_distortion import DetectorDistortionBase, DetectorDistortionMapsXZ
import silx.io


def DetectorDistortionProvider(detector_full_shape_vh=(0, 0), correction_type="", options=""):
    if correction_type == "identity":
        return DetectorDistortionBase(detector_full_shape_vh=detector_full_shape_vh)
    elif correction_type == "map_xz":
        options = options.replace("path=", "path_eq")
        user_params = extract_parameters(options)
        print(user_params, options)
        map_x = silx.io.get_data(user_params["map_x"].replace("path_eq", "path="))
        map_z = silx.io.get_data(user_params["map_z"].replace("path_eq", "path="))
        return DetectorDistortionMapsXZ(map_x=map_x, map_z=map_z)
    else:
        message = f"""
        Unknown correction type: {correction_type} requested
        """
        raise ValueError(message)



import os
import pytest
import numpy
from tqdm import tqdm

from nabu.stitching.y_stitching import y_stitching
from nabu.stitching.config import PreProcessedYStitchingConfiguration
from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan


def build_nxtomos(output_dir, flip_lr, flip_ud) -> tuple:
    r"""
    build two nxtomos in output_dir and return the list of NXtomos ready to be stitched
      /\
      |        ______________       ______________
      |       |~           ~~|      |~             |
      |       |~ nxtomo 1  ~~|      |~ nxtomo 0    |
    Z |       |~ frame     ~~|      |~ frame       |
              |______________|      |______________|
    <-----------------------------------------------
                    90                     40      0
                    y (in acquisition space)
    * ~: represent the overlap area
    """
    dark_data = numpy.array([0] * 64 * 120, dtype=numpy.float32).reshape((64, 120))
    flat_data = numpy.array([1] * 64 * 120, dtype=numpy.float32).reshape((64, 120))
    normalized_data = numpy.linspace(128, 1024, num=64 * 120, dtype=numpy.float32).reshape((64, 120))
    if flip_lr:
        dark_data = numpy.fliplr(dark_data)
        flat_data = numpy.fliplr(flat_data)
        normalized_data = numpy.fliplr(normalized_data)
    if flip_ud:
        dark_data = numpy.flipud(dark_data)
        flat_data = numpy.flipud(flat_data)
        normalized_data = numpy.flipud(normalized_data)

    raw_data = (normalized_data + dark_data) * (flat_data + dark_data)

    # create raw data
    scans = []
    slices = (slice(0, 80), slice(60, -1))
    frame_y_positions = (40, 90)
    for i_nxtomo, (my_slice, frame_y_position) in enumerate(zip(slices, frame_y_positions)):
        my_raw_data = raw_data[:, my_slice]
        assert my_raw_data.ndim == 2
        my_dark_data = dark_data[:, my_slice]
        assert my_dark_data.ndim == 2
        my_flat_data = flat_data[:, my_slice]
        assert my_flat_data.ndim == 2

        n_projs = 3
        nx_tomo = NXtomo()
        nx_tomo.sample.x_translation = [0] * (n_projs + 2)
        nx_tomo.sample.y_translation = [frame_y_position] * (n_projs + 2)
        nx_tomo.sample.z_translation = [0] * (n_projs + 2)
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=(n_projs + 2), endpoint=False)
        nx_tomo.instrument.detector.image_key_control = (
            ImageKey.DARK_FIELD,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
        )
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = numpy.stack(
            (
                my_dark_data,
                my_flat_data,
                my_raw_data,
                my_raw_data,
                my_raw_data,
            )
        )

        file_path = os.path.join(output_dir, f"nxtomo_{i_nxtomo}.nx")
        entry = f"entry000{i_nxtomo}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(NXtomoScan(scan=file_path, entry=entry))
    return scans, frame_y_positions, normalized_data


@pytest.mark.parametrize("flip_lr", (True, False))
@pytest.mark.parametrize("flip_ud", (True, False))
@pytest.mark.parametrize("progress", (None, "with_tqdm"))
def test_preprocessing_stitching(tmp_path, flip_lr, flip_ud, progress):
    if progress == "with_tqdm":
        progress = tqdm(total=100)

    nxtomo_dir = tmp_path / "nxtomos"
    nxtomo_dir.mkdir()
    output_dir = tmp_path / "output"
    output_dir.mkdir()

    output_file_path = os.path.join(output_dir, "nxtomo.nxs")

    nxtomos, _, normalized_data = build_nxtomos(
        output_dir=nxtomo_dir,
        flip_lr=flip_lr,
        flip_ud=flip_ud,
    )

    configuration = PreProcessedYStitchingConfiguration(
        input_scans=nxtomos,
        axis_0_pos_px=None,
        axis_1_pos_px=None,
        axis_2_pos_px=None,
        output_file_path=output_file_path,
        output_data_path="stitched_volume",
    )

    output_identifier = y_stitching(
        configuration=configuration,
        progress=progress,
    )
    created_nx_tomo = NXtomo().load(
        file_path=output_identifier.file_path,
        data_path=output_identifier.data_path,
        detector_data_as="as_numpy_array",
    )
    assert created_nx_tomo.instrument.detector.data.shape == (
        3,
        64,
        120,
    )  # 3 == number of projections, dark and flat will not be exported when doing the stitching
    # TODO: improve me: the relative tolerance is pretty high. This doesn't comes from the algorithm on itself
    # but more on the numerical calculation and the flat field normalization
    numpy.testing.assert_allclose(normalized_data, created_nx_tomo.instrument.detector.data[0], rtol=0.06)



import numpy
import pytest
from nabu.stitching.alignment import align_horizontally, PaddedRawData
from nabu.testutils import get_data


def test_alignment_axis_2():
    """
    test 'align_horizontally' function
    """
    dataset = get_data("chelsea.npz")["data"]  # shape is (300, 451)

    # test if new_width < current_width: should raise an error
    with pytest.raises(ValueError):
        align_horizontally(dataset, alignment="center", new_width=10)

    # test some use cases
    res = align_horizontally(
        dataset,
        alignment="center",
        new_width=600,
        pad_mode="mean",
    )
    assert res.shape == (300, 600)
    numpy.testing.assert_array_almost_equal(res[:, 74:-75], dataset)

    res = align_horizontally(
        dataset,
        alignment="left",
        new_width=600,
        pad_mode="median",
    )
    assert res.shape == (300, 600)
    numpy.testing.assert_array_almost_equal(res[:, :451], dataset)

    res = align_horizontally(
        dataset,
        alignment="right",
        new_width=600,
        pad_mode="reflect",
    )
    assert res.shape == (300, 600)
    numpy.testing.assert_array_almost_equal(res[:, -451:], dataset)


def test_PaddedRawData():
    """
    test PaddedVolume class
    """
    data = numpy.linspace(
        start=0,
        stop=20 * 6 * 3,
        dtype=numpy.int64,
        num=20 * 6 * 3,
    )
    data = data.reshape((3, 6, 20))

    padded_volume = PaddedRawData(data=data, axis_1_pad_width=(4, 1))

    assert padded_volume.shape == (3, 6 + 4 + 1, 20)

    numpy.testing.assert_array_equal(
        padded_volume[:, 0, :],
        numpy.zeros(shape=(3, 1, 20), dtype=numpy.int64),
    )
    numpy.testing.assert_array_equal(
        padded_volume[:, 3, :],
        numpy.zeros(shape=(3, 1, 20), dtype=numpy.int64),
    )
    numpy.testing.assert_array_equal(
        padded_volume[:, 10, :],
        numpy.zeros(shape=(3, 1, 20), dtype=numpy.int64),
    )
    assert padded_volume[:, 3, :].shape == (3, 1, 20)
    numpy.testing.assert_array_equal(
        padded_volume[:, 4, :],
        data[:, 0:1, :],  # TODO: have a look, return a 3D array when a 2D expected...
    )

    with pytest.raises(ValueError):
        padded_volume[:, 40, :]
    with pytest.raises(ValueError):
        padded_volume[:, 5:1, :]

    arrays = (
        numpy.zeros(shape=(3, 4, 20), dtype=numpy.int64),
        data,
        numpy.zeros(shape=(3, 1, 20), dtype=numpy.int64),
    )
    expected_volume = numpy.hstack(
        arrays,
    )
    assert padded_volume[:, :, :].shape == padded_volume.shape
    assert expected_volume.shape == padded_volume.shape

    numpy.testing.assert_array_equal(
        padded_volume[:, :, :],
        expected_volume,
    )



import numpy as np
import pytest
from scipy.ndimage import shift
from nabu.pipeline.params import fbp_filters
from nabu.utils import clip_circle
from nabu.testutils import get_data, generate_tests_scenarios, __do_long_tests__
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import get_opencl_context, __has_pyopencl__

from nabu.processing.fft_cuda import has_skcuda, has_vkfft as has_vkfft_cu
from nabu.processing.fft_opencl import has_vkfft as has_vkfft_cl

__has_pycuda__ = __has_pycuda__ and (has_skcuda() or has_vkfft_cu())
__has_pyopencl__ = __has_pyopencl__ and has_vkfft_cl()

if __has_pycuda__:
    from nabu.reconstruction.fbp import CudaBackprojector
    from nabu.reconstruction.hbp import HierarchicalBackprojector
if __has_pyopencl__:
    from nabu.reconstruction.fbp_opencl import OpenCLBackprojector


scenarios = generate_tests_scenarios({"backend": ["cuda", "opencl"]})
if __do_long_tests__:
    scenarios = generate_tests_scenarios(
        {
            "backend": ["cuda", "opencl"],
            "input_on_gpu": [False, True],
            "output_on_gpu": [False, True],
            "use_textures": [True, False],
        }
    )


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino_512 = get_data("mri_sino500.npz")["data"]
    cls.ref_512 = get_data("mri_rec_astra.npz")["data"]
    # always use contiguous arrays
    cls.sino_511 = np.ascontiguousarray(cls.sino_512[:, :-1])
    # Could be set to 5.0e-2 when using textures. When not using textures, interpolation slightly differs
    cls.tol = 5.1e-2

    if __has_pycuda__:
        cls.cuda_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.opencl_ctx = get_opencl_context("all")
    yield
    if __has_pycuda__:
        cls.cuda_ctx.pop()


def clip_to_inner_circle(img, radius_factor=0.99, out_value=0):
    radius = int(radius_factor * max(img.shape) / 2)
    return clip_circle(img, radius=radius, out_value=out_value)


@pytest.mark.usefixtures("bootstrap")
class TestFBP:

    def _get_backprojector(self, config, *bp_args, **bp_kwargs):
        if config["backend"] == "cuda":
            if not (__has_pycuda__):
                pytest.skip("Need pycuda + (scikit-cuda or pyvkfft)")
            Backprojector = CudaBackprojector
            ctx = self.cuda_ctx
        else:
            if not (__has_pyopencl__):
                pytest.skip("Need pyopencl + pyvkfft")
            Backprojector = OpenCLBackprojector
            ctx = self.opencl_ctx
        if config.get("use_textures", True) is False:
            # patch "extra_options"
            extra_options = bp_kwargs.pop("extra_options", {})
            extra_options["use_textures"] = False
            bp_kwargs["extra_options"] = extra_options
        return Backprojector(*bp_args, **bp_kwargs, backend_options={"ctx": ctx})

    @staticmethod
    def apply_fbp(config, backprojector, sinogram):
        if config.get("input_on_gpu", False):
            sinogram = backprojector._processing.set_array("sinogram", sinogram)
        if config.get("output_on_gpu", False):
            output = backprojector._processing.allocate_array("output", backprojector.slice_shape, dtype="f")
        else:
            output = None
        res = backprojector.fbp(sinogram, output=output)
        if config.get("output_on_gpu", False):
            res = res.get()
        return res

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_512(self, config):
        """
        Simple test of a FBP on a 512x512 slice
        """
        B = self._get_backprojector(config, (500, 512))
        res = self.apply_fbp(config, B, self.sino_512)

        delta_clipped = clip_to_inner_circle(res - self.ref_512)
        err_max = np.max(np.abs(delta_clipped))

        assert err_max < self.tol, "Something wrong with config=%s" % (str(config))

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_511(self, config):
        """
        Test FBP of a 511x511 slice where the rotation axis is at (512-1)/2.0
        """
        B = self._get_backprojector(config, (500, 511), rot_center=255.5)
        res = self.apply_fbp(config, B, self.sino_511)
        ref = self.ref_512[:-1, :-1]

        delta_clipped = clip_to_inner_circle(res - ref)
        err_max = np.max(np.abs(delta_clipped))

        assert err_max < self.tol, "Something wrong with config=%s" % (str(config))

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_roi(self, config):
        """
        Test FBP in region of interest
        """
        sino = self.sino_511

        B0 = self._get_backprojector(config, sino.shape, rot_center=255.5)
        ref = B0.fbp(sino)

        def backproject_roi(roi, reference):
            B = self._get_backprojector(config, sino.shape, rot_center=255.5, slice_roi=roi)
            res = self.apply_fbp(config, B, sino)
            err_max = np.max(np.abs(res - reference))
            return err_max

        cases = {
            # Test 1: use slice_roi=(0, -1, 0, -1), i.e plain FBP of whole slice
            1: [(0, None, 0, None), ref],
            # Test 2: horizontal strip
            2: [(0, None, 50, 55), ref[50:55, :]],
            # Test 3: vertical strip
            3: [(60, 65, 0, None), ref[:, 60:65]],
            # Test 4: rectangular inner ROI
            4: [(157, 162, 260, -10), ref[260:-10, 157:162]],
        }
        for roi, ref in cases.values():
            err_max = backproject_roi(roi, ref)
            assert err_max < self.tol, "Something wrong with ROI = %s for config=%s" % (
                str(roi),
                str(config),
            )

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_axis_corr(self, config):
        """
        Test the "axis correction" feature
        """
        sino = self.sino_512

        # Create a sinogram with a drift in the rotation axis
        def create_drifted_sino(sino, drifts):
            out = np.zeros_like(sino)
            for i in range(sino.shape[0]):
                out[i] = shift(sino[i], drifts[i])
            return out

        drifts = np.linspace(0, 20, sino.shape[0])
        sino = create_drifted_sino(sino, drifts)

        B = self._get_backprojector(config, sino.shape, extra_options={"axis_correction": drifts})
        res = self.apply_fbp(config, B, sino)

        delta_clipped = clip_circle(res - self.ref_512, radius=200)
        err_max = np.max(np.abs(delta_clipped))
        # Max error is relatively high, migh be due to interpolation of scipy shift in sinogram
        assert err_max < 10.0, "Max error is too high"

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_clip_circle(self, config):
        """
        Test the "clip outer circle" parameter in (extra options)
        """
        sino = self.sino_512
        tol = 1e-5

        for rot_center in [None, sino.shape[1] / 2.0 - 10, sino.shape[1] / 2.0 + 15]:
            B = self._get_backprojector(
                config, sino.shape, rot_center=rot_center, extra_options={"clip_outer_circle": True}
            )
            res = self.apply_fbp(config, B, sino)

            B0 = self._get_backprojector(
                config, sino.shape, rot_center=rot_center, extra_options={"clip_outer_circle": False}
            )
            res_noclip = B0.fbp(sino)
            ref = clip_to_inner_circle(res_noclip, radius_factor=1)
            abs_diff = np.abs(res - ref)
            err_max = np.max(abs_diff)
            assert err_max < tol, "Max error is too high for rot_center=%s ; %s" % (str(rot_center), str(config))

            # Test with custom outer circle value
            B1 = self._get_backprojector(
                config,
                sino.shape,
                rot_center=rot_center,
                extra_options={"clip_outer_circle": True, "outer_circle_value": np.nan},
            )
            res1 = self.apply_fbp(config, B1, sino)
            ref1 = clip_to_inner_circle(res_noclip, radius_factor=1, out_value=np.nan)
            abs_diff1 = np.abs(res1 - ref1)
            err_max1 = np.nanmax(abs_diff1)
            assert err_max1 < tol, "Max error is too high for rot_center=%s ; %s" % (str(rot_center), str(config))

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_centered_axis(self, config):
        """
        Test the "centered_axis" parameter (in extra options)
        """
        sino = np.pad(self.sino_512, ((0, 0), (100, 0)))
        rot_center = (self.sino_512.shape[1] - 1) / 2.0 + 100

        B0 = self._get_backprojector(config, self.sino_512.shape)
        ref = B0.fbp(self.sino_512)

        # Check that "centered_axis" worked

        B = self._get_backprojector(config, sino.shape, rot_center=rot_center, extra_options={"centered_axis": True})
        res = self.apply_fbp(config, B, sino)
        # The outside region (outer circle) is different as "res" is a wider slice
        diff = clip_to_inner_circle(res[50:-50, 50:-50] - ref)
        err_max = np.max(np.abs(diff))
        assert err_max < 5e-2, "centered_axis without clip_circle: something wrong"

        # Check that "clip_outer_circle" works when used jointly with "centered_axis"
        B = self._get_backprojector(
            config,
            sino.shape,
            rot_center=rot_center,
            extra_options={
                "centered_axis": True,
                "clip_outer_circle": True,
            },
        )
        res2 = self.apply_fbp(config, B, sino)
        diff = res2 - clip_to_inner_circle(res, radius_factor=1)
        err_max = np.max(np.abs(diff))
        assert err_max < 1e-5, "centered_axis with clip_circle: something wrong"

    @pytest.mark.parametrize("config", scenarios)
    def test_fbp_filters(self, config):
        for filter_name in set(fbp_filters.values()):
            if filter_name in [None, "ramlak"]:
                continue
            B = self._get_backprojector(config, self.sino_512.shape, filter_name=filter_name)
            self.apply_fbp(config, B, self.sino_512)
            # not sure what to check in this case

    @pytest.mark.parametrize("config", scenarios)
    def test_differentiated_backprojection(self, config):
        # test Hilbert + DBP
        sino_diff = np.diff(self.sino_512, axis=1, prepend=0).astype("f")
        # Need to translate the axis a little bit, because of non-centered differentiation.
        # prepend -> +0.5 ; append -> -0.5
        B = self._get_backprojector(config, sino_diff.shape, filter_name="hilbert", rot_center=255.5 + 0.5)
        rec = self.apply_fbp(config, B, sino_diff)
        # Looks good, but all frequencies are not recovered. Use a metric like SSIM or FRC ?


@pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for using HBP")
@pytest.mark.usefixtures("bootstrap")
class TestHBP:

    def _compare_to_reference(self, res, ref, err_msg="", radius_factor=0.9, rel_tol=0.02):
        delta_clipped = clip_to_inner_circle(res - ref, radius_factor=radius_factor)
        err_max = np.max(np.abs(delta_clipped))
        err_max_rel = err_max / ref.max()
        assert err_max_rel < rel_tol, err_msg

    def test_hbp_simple(self):
        B = HierarchicalBackprojector(self.sino_512.shape)
        res = B.fbp(self.sino_512)
        self._compare_to_reference(res, self.ref_512)

    def test_hbp_input_output(self):
        B = HierarchicalBackprojector(self.sino_512.shape)

        d_sino = B._processing.to_device("d_sino2", self.sino_512)
        d_slice = B._processing.allocate_array("d_slice2", self.ref_512.shape)
        h_slice = np.zeros_like(self.ref_512)

        # in: host, out: host (not provided)
        # see test above

        # in: host, out: host (provided)
        res = B.fbp(self.sino_512, output=h_slice)
        self._compare_to_reference(h_slice, self.ref_512, err_msg="in: host, out: host (provided)")
        h_slice.fill(0)

        # in: host, out: device
        res = B.fbp(self.sino_512, output=d_slice)
        self._compare_to_reference(d_slice.get(), self.ref_512, err_msg="in: host, out: device")
        d_slice.fill(0)

        # in: device, out: host (not provided)
        res = B.fbp(d_sino)
        self._compare_to_reference(res, self.ref_512, err_msg="in: device, out: host (not provided)")

        # in: device, out: host (provided)
        res = B.fbp(d_sino, output=h_slice)
        self._compare_to_reference(h_slice, self.ref_512, err_msg="in: device, out: host (provided)")
        h_slice.fill(0)

        # in: device, out: device
        res = B.fbp(d_sino, output=d_slice)
        self._compare_to_reference(d_slice.get(), self.ref_512, err_msg="in: device, out: device")
        d_slice.fill(0)

    def test_hbp_cor(self):
        """
        Test HBP with various sinogram shapes, obtained by truncating horizontally the original sinogram.
        The Center of rotation is always 255.5 (the one of original sinogram), so it also tests reconstruction with a shifted CoR.
        """
        for crop in [1, 2, 5, 10]:
            sino = np.ascontiguousarray(self.sino_512[:, :-crop])
            B = HierarchicalBackprojector(sino.shape, rot_center=255.5)
            res = B.fbp(sino)

            # HBP always uses "centered_axis=1", so we cannot compare non-integer shifts
            if crop % 2 == 0:
                ref = self.ref_512[crop // 2 : -crop // 2, crop // 2 : -crop // 2]
                self._compare_to_reference(res, ref, radius_factor=0.95, rel_tol=0.02)

    def test_hbp_clip_circle(self):
        B_clip = HierarchicalBackprojector(self.sino_512.shape, extra_options={"clip_outer_circle": True})
        B_noclip = HierarchicalBackprojector(self.sino_512.shape, extra_options={"clip_outer_circle": False})
        res_clip = B_clip.fbp(self.sino_512)
        res_noclip = B_noclip.fbp(self.sino_512)
        self._compare_to_reference(res_clip, clip_to_inner_circle(res_noclip, radius_factor=1), "clip_circle")

    def test_hbp_axis_corr(self):
        sino = self.sino_512

        # Create a sinogram with a drift in the rotation axis
        def create_drifted_sino(sino, drifts):
            out = np.zeros_like(sino)
            for i in range(sino.shape[0]):
                out[i] = shift(sino[i], drifts[i])
            return out

        drifts = np.linspace(0, 20, sino.shape[0])
        sino = create_drifted_sino(sino, drifts)

        B = HierarchicalBackprojector(sino.shape, extra_options={"axis_correction": drifts})
        res = B.fbp(sino)

        # Max error is relatively high, migh be due to interpolation of scipy shift in sinogram
        self._compare_to_reference(res, self.ref_512, radius_factor=0.95, rel_tol=0.04, err_msg="axis_corr")

    @pytest.mark.skipif(not (__do_long_tests__), reason="need NABU_LONG_TESTS=1 for this test")
    def test_hbp_scale_factor(self):
        scale_factor = 0.03125
        B_scaled = HierarchicalBackprojector(self.sino_512.shape, extra_options={"scale_factor": scale_factor})
        B_unscaled = HierarchicalBackprojector(self.sino_512.shape)
        res_scaled = B_scaled.fbp(self.sino_512)
        res_unscaled = B_unscaled.fbp(self.sino_512)
        self._compare_to_reference(res_scaled, res_unscaled * scale_factor, rel_tol=1e-7, err_msg="scale_factor")



import pytest
import numpy as np
import scipy.ndimage
from scipy import sparse
from nabu.io.detector_distortion import DetectorDistortionBase
from nabu.processing.rotation import Rotation, __have__skimage__

if __have__skimage__:
    import skimage


@pytest.mark.skipif(not (__have__skimage__), reason="Need scikit-image for rotation")
def test_detector_distortion():
    image = scipy.ndimage.gaussian_filter(np.random.random([379, 1357]), 3.0)
    center_xz = ((image.shape[1] - 1) / 2, (image.shape[0] - 1) / 2)

    part_to_be_retrieved = image[100:279]

    rotated_image = skimage.transform.rotate(image, angle=5.0, center=center_xz[::])

    corrector = DetectorDistortionRotation(detector_full_shape_vh=image.shape, center_xz=center_xz, angle_deg=-5.0)
    start_x, end_x, start_z, end_z = corrector.set_sub_region_transformation(
        target_sub_region=(
            None,
            None,
            100,
            279,
        )
    )

    source = rotated_image[start_z:end_z, start_x:end_x]

    retrieved = corrector.transform(source)

    diff = (retrieved - part_to_be_retrieved)[:, 20:-20]

    assert abs(diff).std() < 1e-3


class DetectorDistortionRotation(DetectorDistortionBase):
    """ """

    def __init__(self, detector_full_shape_vh=(0, 0), center_xz=(0, 0), angle_deg=0.0):
        """This is the basis class.
        A simple identity transformation which has the only merit to show
        how it works.Reimplement this function to have more parameters for other
        transformations
        """
        self._build_full_transformation(detector_full_shape_vh, center_xz, angle_deg)

    def _build_full_transformation(self, detector_full_shape_vh, center_xz, angle_deg):
        """A simple identity.
        Reimplement this function to have more parameters for other
        transformations
        """

        indices = np.indices(detector_full_shape_vh)
        center_x, center_z = center_xz
        coordinates = (indices.T - np.array([center_z, center_x])).T

        c = np.cos(np.deg2rad(angle_deg))
        s = np.sin(np.deg2rad(angle_deg))

        rot_mat = np.array([[c, s], [-s, c]])

        coordinates = np.tensordot(rot_mat, coordinates, axes=[1, 0])

        # padding
        sz, sx = detector_full_shape_vh
        total_detector_npixs = sz * sx
        xs = np.clip(np.array(coordinates[1].flat) + center_x, [[0]], [[sx - 1]])
        zs = np.clip(np.array(coordinates[0].flat) + center_z, [[0]], [[sz - 1]])

        ix0s = np.floor(xs)
        ix1s = np.ceil(xs)
        fx = xs - ix0s

        iz0s = np.floor(zs)
        iz1s = np.ceil(zs)
        fz = zs - iz0s

        I_tmp = np.empty([4 * sz * sx], np.int64)
        J_tmp = np.empty([4 * sz * sx], np.int64)
        V_tmp = np.ones([4 * sz * sx], "f")

        I_tmp[:] = np.arange(sz * sx * 4) // 4

        J_tmp[0::4] = iz0s * sx + ix0s
        J_tmp[1::4] = iz0s * sx + ix1s
        J_tmp[2::4] = iz1s * sx + ix0s
        J_tmp[3::4] = iz1s * sx + ix1s

        V_tmp[0::4] = (1 - fz) * (1 - fx)
        V_tmp[1::4] = (1 - fz) * fx
        V_tmp[2::4] = fz * (1 - fx)
        V_tmp[3::4] = fz * fx

        self.detector_full_shape_vh = detector_full_shape_vh

        coo_tmp = sparse.coo_matrix((V_tmp.astype("f"), (I_tmp, J_tmp)), shape=(sz * sx, sz * sx))

        csr_tmp = coo_tmp.tocsr()

        self.full_csr_data = csr_tmp.data
        self.full_csr_indices = csr_tmp.indices
        self.full_csr_indptr = csr_tmp.indptr

        ## This will be used to save time if the same sub_region argument is requested several time in a row
        self._status = None

    def _set_sub_region_transformation(
        self,
        target_sub_region=(
            (
                None,
                None,
                0,
                0,
            ),
        ),
    ):
        (x_start, x_end, z_start, z_end) = target_sub_region

        if z_start is None:
            z_start = 0
        if z_end is None:
            z_end = self.detector_full_shape_vh[0]

        if (x_start, x_end) not in [(None, None), (0, None), (0, self.detector_full_shape_vh[1])]:
            message = f""" In the base class DetectorDistortionRotation only vertical slicing is accepted.
            The sub_region contained (x_start, x_end)={(x_start, x_end)} which would slice the 
            full horizontal size which is {self.detector_full_shape_vh[1]}
            """
            raise ValueError()

        x_start, x_end = 0, self.detector_full_shape_vh[1]

        row_ptr_start = z_start * self.detector_full_shape_vh[1]
        row_ptr_end = z_end * self.detector_full_shape_vh[1]

        indices_start = self.full_csr_indptr[row_ptr_start]
        indices_end = self.full_csr_indptr[row_ptr_end]

        data_tmp = self.full_csr_data[indices_start:indices_end]

        target_offset = self.full_csr_indptr[row_ptr_start]
        indptr_tmp = self.full_csr_indptr[row_ptr_start : row_ptr_end + 1] - target_offset

        indices_tmp = self.full_csr_indices[indices_start:indices_end]

        iz_source = (indices_tmp) // self.detector_full_shape_vh[1]

        z_start_source = iz_source.min()
        z_end_source = iz_source.max() + 1
        source_offset = z_start_source * self.detector_full_shape_vh[1]
        indices_tmp = indices_tmp - source_offset

        target_size = (z_end - z_start) * self.detector_full_shape_vh[1]
        source_size = (z_end_source - z_start_source) * self.detector_full_shape_vh[1]

        self.transformation_matrix = sparse.csr_matrix(
            (data_tmp, indices_tmp, indptr_tmp), shape=(target_size, source_size)
        )

        self.target_shape = ((z_end - z_start), self.detector_full_shape_vh[1])

        ## For the identity matrix the source and the target have the same size.
        ## The two following lines are trivial.
        ## For this identity transformation only the slicing of the appropriate part
        ## of the identity sparse matrix is slightly laborious.
        ## Practical case will be more complicated and source_sub_region
        ## will be in general larger than the target_sub_region
        self._status = {
            "target_sub_region": ((x_start, x_end, z_start, z_end)),
            "source_sub_region": ((x_start, x_end, z_start_source, z_end_source)),
        }

        return self._status["source_sub_region"]



import os
from ..resources.logger import LoggerOrPrint
from ..utils import copy_dict_items
from ..reconstruction.sinogram import get_extended_sinogram_width


class DatasetValidatorBase:
    # this in the helical derived class will be False
    _check_also_z = True

    def __init__(self, nabu_config, dataset_info, logger=None):
        """
        Perform a coupled validation of nabu configuration against dataset information.
        Check the consistency of these two structures, and modify them in-place.

        Parameters
        ----------
        nabu_config: dict
            Dictionary containing the nabu configuration, usually got from
            `nabu.pipeline.config.validate_config()`
            It will be modified !
        dataset_info: `DatasetAnalyzer` instance
            Structure containing information on the dataset to process.
            It will be modified !
        """
        self.nabu_config = nabu_config
        self.dataset_info = dataset_info
        self.logger = LoggerOrPrint(logger)
        self.rec_params = copy_dict_items(self.nabu_config["reconstruction"], self.nabu_config["reconstruction"].keys())
        self._validate()

    def _validate(self):
        raise ValueError("Base class")

    @property
    def is_halftomo(self):
        do_halftomo = self.nabu_config["reconstruction"].get("enable_halftomo", False)
        if do_halftomo == "auto":
            do_halftomo = self.dataset_info.is_halftomo
            if do_halftomo is None:
                raise ValueError(
                    "'enable_halftomo' was set to 'auto' but unable to get the information on field of view"
                )
        return do_halftomo

    def _check_not_empty(self):
        if len(self.dataset_info.projections) == 0:
            msg = "Dataset seems to be empty (no projections)"
            self.logger.fatal(msg)
            raise ValueError(msg)
        if self.dataset_info.n_angles is None:
            msg = "Could not determine the number of projections. Please check the .info or HDF5 file"
            self.logger.fatal(msg)
            raise ValueError(msg)
        for dim_name, n in zip(["dim_1", "dim_2"], self.dataset_info.radio_dims):
            if n is None:
                msg = "Could not determine %s. Please check the .info file or HDF5 file" % dim_name
                self.logger.fatal(msg)
                raise ValueError(msg)

    @staticmethod
    def _convert_negative_idx(idx, last_idx):
        res = idx
        if idx < 0:
            res = last_idx + idx
        return res

    def _get_nx_ny(self, binning_factor=1):
        nx = self.dataset_info.radio_dims[0] // binning_factor
        if self.is_halftomo:
            cor = self._get_cor(binning_factor=binning_factor)
            nx = get_extended_sinogram_width(nx, cor)
        ny = nx
        return nx, ny

    def _get_cor(self, binning_factor=1):
        cor = self.dataset_info.axis_position
        if binning_factor >= 1:
            # Backprojector uses middle of pixel for coordinate indices.
            # This means that the leftmost edge of the leftmost pixel has coordinate -0.5.
            # When using binning with a factor 'b', the CoR has to adapted as
            #   cor_binned  =  (cor + 0.5)/b - 0.5
            cor = (cor + 0.5) / binning_factor - 0.5
        return cor

    def _convert_negative_indices(self):
        """
        Convert any negative index to the corresponding positive index.
        """
        nx, nz = self.dataset_info.radio_dims
        ny = nx
        if self.is_halftomo:
            if self.dataset_info.axis_position is None:
                raise ValueError(
                    "Cannot use rotation axis position in the middle of the detector when half tomo is enabled"
                )
            nx, ny = self._get_nx_ny()

        what = (
            ("start_x", nx),
            ("end_x", nx),
            ("start_y", ny),
            ("end_y", ny),
        )
        if self._check_also_z:
            what = what + (
                ("start_z", nz),
                ("end_z", nz),
            )
        for key, upper_bound in what:
            val = self.rec_params[key]
            if isinstance(val, str):
                idx_mapping = {
                    "first": 0,
                    "middle": upper_bound // 2,  # works on both start_ and end_ since the end_ index is included
                    "last": upper_bound - 1,  # upper bound is included in the user interface (contrarily to python)
                }
                res = idx_mapping[val]
            else:
                res = self._convert_negative_idx(self.rec_params[key], upper_bound)
            self.rec_params[key] = res
            self.rec_region = copy_dict_items(self.rec_params, [w[0] for w in what])

    def _get_output_filename(self):
        # This function modifies nabu_config !
        opts = self.nabu_config["output"]
        dataset_path = self.nabu_config["dataset"]["location"]
        if opts["location"] == "" or opts["location"] is None:
            opts["location"] = os.path.dirname(dataset_path)
        if opts["file_prefix"] == "" or opts["file_prefix"] is None:
            if os.path.isfile(dataset_path):  # hdf5
                file_prefix = os.path.basename(dataset_path).split(".")[0]
            elif os.path.isdir(dataset_path):
                file_prefix = os.path.basename(dataset_path)
            else:
                raise ValueError("dataset location %s is neither a file or directory" % dataset_path)
            file_prefix += "_rec"  # avoid overwriting dataset
            opts["file_prefix"] = file_prefix

    @staticmethod
    def _check_start_end_idx(start, end, n_elements, start_name="start_x", end_name="end_x"):
        assert start >= 0 and start < n_elements, "Invalid value %d for %s, must be >= 0 and < %d" % (
            start,
            start_name,
            n_elements,
        )
        assert end >= 0 and end < n_elements, "Invalid value for %d %s, must be >= 0 and < %d" % (
            end,
            end_name,
            n_elements,
        )
        assert start <= end, "Must have %s <= %s" % (start_name, end_name)

    def _handle_binning(self):
        """
        Modify the dataset description/process config to handle binning and projections subsampling
        """
        dataset_cfg = self.nabu_config["dataset"]
        self.binning = (dataset_cfg["binning"], dataset_cfg["binning_z"])
        subsampling_factor, subsampling_start = dataset_cfg["projections_subsampling"]
        self.subsampling_factor = subsampling_factor or 1
        self.subsampling_start = subsampling_start or 0

        if self.binning != (1, 1):
            bin_x, bin_z = self.binning
            rec_cfg = self.rec_params

            # Update "start_xyz"
            rec_cfg["start_x"] //= bin_x
            rec_cfg["start_y"] //= bin_x
            rec_cfg["start_z"] //= bin_z

            # Update "end_xyz". Things are a little bit more complicated for several reasons:
            #   - In the user interface (configuration file), end_xyz index is INCLUDED (contrarily to python). So there are +1, -1 all over the place.
            #   - When using half tomography, n_x and n_y are less straightforward : 2*CoR(binning) instead of 2*CoR//binning
            #   - delta = end - start [+1]  should be a multiple of binning factor. This makes things much easier for processing pipeline.

            def ensure_multiple_of_binning(end, start, binning_factor):
                """
                Update "end" so that end-start is a multiple of "binning_factor"
                Note that "end" is INCLUDED here (comes from user configuration)
                """
                return end - ((end - start + 1) % binning_factor)

            end_z = ensure_multiple_of_binning(rec_cfg["end_z"], rec_cfg["start_z"], bin_z)
            rec_cfg["end_z"] = (end_z + 1) // bin_z - 1

            nx_binned, ny_binned = self._get_nx_ny(binning_factor=bin_x)

            end_y = ensure_multiple_of_binning(rec_cfg["end_y"], rec_cfg["start_y"], bin_x)
            rec_cfg["end_y"] = min((end_y + 1) // bin_x - 1, ny_binned - 1)

            end_x = ensure_multiple_of_binning(rec_cfg["end_x"], rec_cfg["start_x"], bin_x)
            rec_cfg["end_x"] = min((end_x + 1) // bin_x - 1, nx_binned - 1)

    def _check_output_file(self):
        out_cfg = self.nabu_config["output"]
        out_fname = os.path.join(out_cfg["location"], out_cfg["file_prefix"] + out_cfg["file_format"])
        if os.path.exists(out_fname):
            raise ValueError("File %s already exists" % out_fname)

    def _handle_processing_mode(self):
        mode = self.nabu_config["resources"]["method"]
        if mode == "preview":
            print(
                "Warning: the method 'preview' was selected. This means that the data volume will be binned so that everything fits in memory."
            )
            # TODO automatically compute binning/subsampling factors as a function of lowest memory (GPU)
            self.nabu_config["dataset"]["binning"] = 2
            self.nabu_config["dataset"]["binning_z"] = 2
            self.nabu_config["dataset"]["projections_subsampling"] = 2, 0
        # TODO handle other modes



from ..processing.unsharp_opencl import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.unsharp_opencl has been moved to nabu.processing.unsharp_opencl",
    do_print=True,
    func_name="unsharp_opencl",
)



#!/usr/bin/env python
import os
import tarfile
import requests

MATHJAX_PATH = "_static/javascript/MathJax-3.0.5"
MATHJAX_URL = "http://www.silx.org/pub/nabu/static/MathJax-3.0.5.tar.lzma"


def download_file(file_url, target_file_path):
    print("Downloading %s" % file_url)
    rep = requests.get(file_url)
    with open(target_file_path, "wb") as f:
        f.write(rep.content)


def uncompress_file(compressed_file_path, target_directory):
    print("Uncompressing %s into %s" % (compressed_file_path, target_directory))
    with tarfile.open(compressed_file_path) as f:
        f.extractall(path=target_directory)


def main():
    doc_path = os.path.dirname(os.path.realpath(__file__))
    mathjax_path = os.path.join(doc_path, MATHJAX_PATH)
    if os.path.isdir(mathjax_path):
        return
    mathjax_comp_file = mathjax_path + ".tar.lzma"
    download_file(MATHJAX_URL, mathjax_comp_file)
    uncompress_file(mathjax_comp_file, os.path.dirname(mathjax_path))


if __name__ == "__main__":
    main()



import numpy as np
from ..utils import BaseClassError, get_opencl_srcfile, updiv
from ..opencl.kernel import OpenCLKernel
from ..opencl.processing import OpenCLProcessing
from pyopencl.tools import dtype_to_ctype as cl_dtype_to_ctype


class FFTshiftBase:
    KernelCls = BaseClassError
    ProcessingCls = BaseClassError
    dtype_to_ctype = BaseClassError
    backend = "none"

    def __init__(self, shape, dtype, dst_dtype=None, axes=None, **backend_options):
        """

        Parameters
        ----------
        shape: tuple
            Array shape - can be 1D or 2D. 3D is not supported.
        dtype: str or numpy.dtype
            Data type, eg. "f", numpy.complex64, ...
        dst_dtype: str or numpy.dtype
            Output data type. If not provided (default), the shift is done in-place.
        axes: tuple, optional
            Axes over which to shift.  Default is None, which shifts all axes.

        Other parameters
        ----------------
        backend_options:
            named arguments to pass to CudaProcessing or OpenCLProcessing
        """
        #
        if axes not in [1, (1,), (-1,)]:
            raise NotImplementedError
        #
        self.processing = self.ProcessingCls(**backend_options)
        self.shape = shape
        if len(self.shape) not in [1, 2]:
            raise ValueError("Expected 1D or 2D array")
        self.dtype = np.dtype(dtype)
        self.dst_dtype = dst_dtype

        if dst_dtype is None:
            self._configure_inplace_shift()
        else:
            self._configure_out_of_place_shift()
        self._configure_kenel_initialization()
        self._fftshift_kernel = self.KernelCls(*self._kernel_init_args, **self._kernel_init_kwargs)
        self._configure_kernel_call()

    def _configure_inplace_shift(self):
        self.inplace = True
        # in-place on odd-sized array is more difficult - see fftshift.cl
        if self.shape[-1] & 1:
            raise NotImplementedError
        #
        self._kernel_init_args = [
            "fftshift_x_inplace",
        ]
        self._kernel_init_kwargs = {
            "options": [
                "-DDTYPE=%s" % self.dtype_to_ctype(self.dtype),
            ],
        }

    def _configure_out_of_place_shift(self):
        self.inplace = False
        self._kernel_init_args = [
            "fftshift_x",
        ]
        self._kernel_init_kwargs = {
            "options": [
                "-DDTYPE=%s" % self.dtype_to_ctype(self.dtype),
                "-DDTYPE_OUT=%s" % self.dtype_to_ctype(np.dtype(self.dst_dtype)),
            ],
        }
        additional_flag = None
        input_is_complex = np.iscomplexobj(np.ones(1, dtype=self.dtype))
        output_is_complex = np.iscomplexobj(np.ones(1, dtype=self.dst_dtype))
        if not (input_is_complex) and output_is_complex:
            additional_flag = "-DCAST_TO_COMPLEX"
        if input_is_complex and not (output_is_complex):
            additional_flag = "-DCAST_TO_REAL"
        if additional_flag is not None:
            self._kernel_init_kwargs["options"].append(additional_flag)

    def _call_fftshift_inplace(self, arr, direction):
        self._fftshift_kernel(  # pylint: disable=E1102
            arr, np.int32(self.shape[1]), np.int32(self.shape[0]), np.int32(direction), **self._kernel_kwargs
        )
        return arr

    def _call_fftshift_out_of_place(self, arr, dst, direction):
        if dst is None:
            dst = self.processing.allocate_array("dst", arr.shape, dtype=self.dst_dtype)
        self._fftshift_kernel(  # pylint: disable=E1102
            arr, dst, np.int32(self.shape[1]), np.int32(self.shape[0]), np.int32(direction), **self._kernel_kwargs
        )
        return dst

    def fftshift(self, arr, dst=None):
        if self.inplace:
            return self._call_fftshift_inplace(arr, 1)
        else:
            return self._call_fftshift_out_of_place(arr, dst, 1)

    def ifftshift(self, arr, dst=None):
        if self.inplace:
            return self._call_fftshift_inplace(arr, -1)
        else:
            return self._call_fftshift_out_of_place(arr, dst, -1)


class OpenCLFFTshift(FFTshiftBase):
    KernelCls = OpenCLKernel
    ProcessingCls = OpenCLProcessing
    dtype_to_ctype = cl_dtype_to_ctype
    backend = "opencl"

    def _configure_kenel_initialization(self):
        self._kernel_init_args.append(self.processing.ctx)
        self._kernel_init_kwargs.update(
            {
                "filename": get_opencl_srcfile("fftshift.cl"),
                "queue": self.processing.queue,
            }
        )

    def _configure_kernel_call(self):
        # TODO in-place fftshift needs to launch only arr.size//2 threads
        block = (16, 16, 1)
        grid = [updiv(a, b) * b for a, b in zip(self.shape[::-1], block)]
        self._kernel_kwargs = {"global_size": grid, "local_size": block}



from .y_stitching import y_stitching
from .z_stitching import z_stitching
from tomoscan.identifier import BaseIdentifier
from nabu.stitching.config import (
    SingleAxisStitchingConfiguration,
    PreProcessedYStitchingConfiguration,
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
)


def stitching(configuration: SingleAxisStitchingConfiguration, progress=None) -> BaseIdentifier:
    """
    Apply stitching from provided configuration.
    Stitching will be applied along a single axis at the moment.

    like:
                    axis 0
                      ^
                      |
    x-ray             |
    -------->          ------> axis 2
                     /
                    /
                     axis 1
    """
    if isinstance(configuration, (PreProcessedYStitchingConfiguration,)):
        return y_stitching(configuration=configuration, progress=progress)
    elif isinstance(configuration, (PreProcessedZStitchingConfiguration, PostProcessedZStitchingConfiguration)):
        return z_stitching(configuration=configuration, progress=progress)
    else:
        raise NotImplementedError(f"configuration type ({type(configuration)}) is not handled")



import os
import warnings
from multiprocessing import get_context
from multiprocessing.pool import Pool
import numpy as np
from ..utils import check_supported
from .fft_base import _BaseFFT, _BaseVKFFT

try:
    from pyvkfft.cuda import VkFFTApp as vk_cufft

    __has_vkfft__ = True
except (ImportError, OSError):
    __has_vkfft__ = False
    vk_cufft = None
from ..cuda.processing import CudaProcessing

Plan = None
cu_fft = None
cu_ifft = None
__has_skcuda__ = None


def init_skcuda():
    # This needs to be done here, because scikit-cuda creates a Cuda context at import,
    # which can mess things up in some cases.
    # Ugly solution to an ugly problem.
    global __has_skcuda__, Plan, cu_fft, cu_ifft
    try:
        from skcuda.fft import Plan
        from skcuda.fft import fft as cu_fft
        from skcuda.fft import ifft as cu_ifft

        __has_skcuda__ = True
    except ImportError:
        __has_skcuda__ = False


class SKCUFFT(_BaseFFT):
    implem = "skcuda"
    backend = "cuda"
    ProcessingCls = CudaProcessing

    def _configure_batched_transform(self):
        if __has_skcuda__ is None:
            init_skcuda()
        if not (__has_skcuda__):
            raise ImportError("Please install pycuda and scikit-cuda to use the CUDA back-end")

        self.cufft_batch_size = 1
        self.cufft_shape = self.shape
        self._cufft_plan_kwargs = {}
        if (self.axes is not None) and (len(self.axes) < len(self.shape)):
            # In the easiest case, the transform is computed along the fastest dimensions:
            #  - 1D transforms of lines of 2D data
            #  - 2D transforms of images of 3D data (stacked along slow dim)
            #  - 1D transforms of 3D data along fastest dim
            # Otherwise, we have to configure cuda "advanced memory layout".
            data_ndims = len(self.shape)

            if data_ndims == 2:
                n_y, n_x = self.shape
                along_fast_dim = self.axes[0] == 1
                self.cufft_shape = n_x if along_fast_dim else n_y
                self.cufft_batch_size = n_y if along_fast_dim else n_x
                if not (along_fast_dim):
                    # Batched vertical 1D FFT on 2D data need advanced data layout
                    # http://docs.nvidia.com/cuda/cufft/#advanced-data-layout
                    self._cufft_plan_kwargs = {
                        "inembed": np.int32([0]),
                        "istride": n_x,
                        "idist": 1,
                        "onembed": np.int32([0]),
                        "ostride": n_x,
                        "odist": 1,
                    }

            if data_ndims == 3:
                # TODO/FIXME - the following work for C2C but not R2C ?!
                # fast_axes = [(1, 2), (2, 1), (2,)]
                fast_axes = [(2,)]
                if self.axes not in fast_axes:
                    raise NotImplementedError(
                        "With the CUDA backend, batched transform on 3D data is only supported along fastest dimensions"
                    )
                self.cufft_batch_size = self.shape[0]
                self.cufft_shape = self.shape[1:]
                if len(self.axes) == 1:
                    # 1D transform on 3D data: here only supported along fast dim, so batch_size is Nx*Ny
                    self.cufft_batch_size = np.prod(self.shape[:2])
                    self.cufft_shape = (self.shape[-1],)
                if len(self.cufft_shape) == 1:
                    self.cufft_shape = self.cufft_shape[0]

    def _configure_normalization(self, normalize):
        self.normalize = normalize
        if self.normalize == "ortho":
            # TODO
            raise NotImplementedError("Normalization mode 'ortho' is not implemented with CUDA backend yet.")
        self.cufft_scale_inverse = self.normalize == "rescale"

    def _compute_fft_plans(self):
        self.plan_forward = Plan(  # pylint: disable = E1102
            self.cufft_shape,
            self.dtype,
            self.dtype_out,
            batch=self.cufft_batch_size,
            stream=self.processing.stream,
            **self._cufft_plan_kwargs,
            # cufft extensible plan API is only supported after 0.5.1
            # (commit 65288d28ca0b93e1234133f8d460dc6becb65121)
            # but there is still no official 0.5.2
            # ~ auto_allocate=True # cufft extensible plan API
        )
        self.plan_inverse = Plan(  # pylint: disable = E1102
            self.cufft_shape,  # not shape_out
            self.dtype_out,
            self.dtype,
            batch=self.cufft_batch_size,
            stream=self.processing.stream,
            **self._cufft_plan_kwargs,
            # cufft extensible plan API is only supported after 0.5.1
            # (commit 65288d28ca0b93e1234133f8d460dc6becb65121)
            # but there is still no official 0.5.2
            # ~ auto_allocate=True
        )

    def fft(self, array, output=None):
        if output is None:
            output = self.output_fft = self.processing.allocate_array(
                "output_fft", self.shape_out, dtype=self.dtype_out
            )
        cu_fft(array, output, self.plan_forward, scale=False)  # pylint: disable = E1102
        return output

    def ifft(self, array, output=None):
        if output is None:
            output = self.output_ifft = self.processing.allocate_array("output_ifft", self.shape, dtype=self.dtype)
        cu_ifft(  # pylint: disable = E1102
            array,
            output,
            self.plan_inverse,
            scale=self.cufft_scale_inverse,
        )
        return output


class VKCUFFT(_BaseVKFFT):
    """
    Cuda FFT, using VKFFT backend
    """

    implem = "vkfft"
    backend = "cuda"
    ProcessingCls = CudaProcessing
    vkffs_cls = vk_cufft

    def _init_backend(self, backend_options):
        super()._init_backend(backend_options)
        self._vkfft_other_init_kwargs = {"stream": self.processing.stream}


def _has_vkfft(x):
    # should be run from within a Process
    try:
        from nabu.processing.fft_cuda import VKCUFFT, __has_vkfft__

        if not __has_vkfft__:
            return False
        vk = VKCUFFT((16,), "f")
        avail = True
    except (ImportError, RuntimeError, OSError, NameError):
        avail = False
    return avail


def has_vkfft(safe=True):
    """
    Determine whether pyvkfft is available.
    For Cuda GPUs, vkfft relies on nvrtc which supports a narrow range of Cuda devices.
    Unfortunately, it's not possible to determine whether vkfft is available before creating a Cuda context.
    So we create a process (from scratch, i.e no fork), do the test within, and exit.
    This function cannot be tested from a notebook/console, a proper entry point has to be created (if __name__ == "__main__").
    """
    if not safe:
        return _has_vkfft(None)
    ctx = get_context("spawn")
    with Pool(1, context=ctx) as p:
        v = p.map(_has_vkfft, [1])[0]
    return v


def _has_skfft(x):
    # should be run from within a Process
    try:
        from nabu.processing.fft_cuda import SKCUFFT

        sk = SKCUFFT((16,), "f")
        avail = True
    except (ImportError, RuntimeError, OSError, NameError):
        avail = False
    return avail


def has_skcuda(safe=True):
    """
    Determine whether scikit-cuda/CUFFT is available.
    Currently, scikit-cuda will create a Cuda context for Cublas, which can mess up the current execution.
    Do it in a separate thread.
    """
    if not safe:
        return _has_skfft(None)
    ctx = get_context("spawn")
    with Pool(1, context=ctx) as p:
        v = p.map(_has_skfft, [1])[0]
    return v


def get_fft_class(backend="vkfft"):
    backends = {
        "scikit-cuda": SKCUFFT,
        "skcuda": SKCUFFT,
        "cufft": SKCUFFT,
        "scikit": SKCUFFT,
        "vkfft": VKCUFFT,
        "pyvkfft": VKCUFFT,
    }

    def get_fft_cls(asked_fft_backend):
        asked_fft_backend = asked_fft_backend.lower()
        check_supported(asked_fft_backend, list(backends.keys()), "Cuda FFT backend name")
        return backends[asked_fft_backend]

    asked_fft_backend_env = os.environ.get("NABU_FFT_BACKEND", "")
    if asked_fft_backend_env != "":
        return get_fft_cls(asked_fft_backend_env)

    avail_fft_implems = get_available_fft_implems()
    if len(avail_fft_implems) == 0:
        raise RuntimeError("Could not any Cuda FFT implementation. Please install either scikit-cuda or pyvkfft")
    if backend not in avail_fft_implems:
        warnings.warn("Could not get FFT backend '%s'" % backend, RuntimeWarning)
        backend = avail_fft_implems[0]

    return get_fft_cls(backend)


def get_available_fft_implems():
    avail_implems = []
    if has_vkfft(safe=True):
        avail_implems.append("vkfft")
    if has_skcuda(safe=True):
        avail_implems.append("skcuda")
    return avail_implems



import os
from multiprocessing import Pool
import sys
import numpy as np
from scipy.ndimage import gaussian_filter
import h5py
from silx.io.dictdump import h5todict
from nxtomo.application.nxtomo import NXtomo
from .. import version
from ..utils import DictToObj, get_available_threads
from .utils import parse_params_values
from .cli_configs import DiagToPixConfig
from ..pipeline.estimators import oversample


"""
  The operations here below rely on diag objects which are found in the result of a nab-helical run with the diag_zpro_run set to a number > 0
  They are found in the configuration section of the nabu output, in several sequential dataset, with hdf5 dataset keys 
  which are 0,1,2.... corresponding to all the z-windows ( chunck) for which we have collected contributions at different angles
  which are nothing else but pieces of ready to use preprocessed radio , and this for different angles.
  In other words redundant contributions conccurring at the same prepocessed radiography but comming from different moment 
  of the helical scan are kept separate
  Forming pairs from contributions for same angle they should coincide where they both have signal ( part of them can be dark if the detector is out of view above or below)

  For each key there is a sequence of radio, the corresponding sequence of weights map, the corresponding z translation, and angles

  The number passed to diag_zpro_run object, is >1, and is interpreted by the diagnostic collection run 
  as the number of wished collecte angles between 0 and 180. Lets call it num_0_180

  The collection is not done here. Here we exploit the result of a previous collection to deduce, looking at the correlations, which correction
  we must bring to the pixel size

  An example of collection is this :
    
   |_____ diagnostics
   |               |  
                  |__ 0
                  |    |_ radios       (4*num_0_180, chunky, chunkx) 
                  |    |
                  |    |_ weights      (4*num_0_180, chunky, chunkx) 
                  |    |
                  |    |_ angles       ( 4*num_0_180,)
                       |
                       |_ searched_rad ( 2*num_0_180,)    these are all the searched angles between 0 and 360 in radians
                       |
                       |_ zmm_trans    (  4*num_0_180,)    the z translation in mm
                       |
                       |_ zpix_transl  (  4*num_0_180,)    the z translation in pix
                       |
                       |_ pixes_size_mm  scalar

"""


def transform_images(diag, ovs):
    """Filter the radios, and oversample them along the vertical line.
    The method in general is similar to the composite cor finding.
    Several overlapping positions are used to match redundant contributions at
    different rotation stages ( theta and theta+360).
    But beforehand it is beneficial to remove low spatial frequencies.
    And we do oversampling on the fly.

    """
    assert len(ovs) == 2, "oversampling must be specified for both vertical and horizontal dimension"

    diag.radios[:] = diag.radios / diag.weights

    diag.radios = [oversample((ima - gaussian_filter(ima, 20, mode="nearest")), ovs) for ima in diag.radios]

    diag.weights = [oversample(ima, ovs) for ima in diag.weights]


def detailed_merit(diag, shift):
    # res will become the merit summed over all the pairs theta, theta+180
    res = 0.0

    # need to account for the weight also. So this will become the used weight for the pairs theta, theta+180
    res_w = 0.0

    ## The following two variables are very important information to be collected.
    ## On the the z translation over a 360 turn
    ## the other is the pixel size in mm.
    ## At the end of the script, the residual shift for perfect correlation
    ## will used to correct zpix_mm, doing a pro-rata with respect to
    ## the z observed translation over one turn
    observed_oneturn_total_shift_zpix_list = []
    zpix_mm = None

    n_angles_2pi = len(diag.radios) // 2
    # In accordance with the collection layout for diagnostics (diag_zpro_run parameter passed to nabu-helical)
    # there are n_angles_pi in [0,180[, and then again the same number of possibly valid radios
    # (check for nan in z translation) in [180,360[, [360,540[, 540,720[
    # So we have len(diag.radios) // 2 in the range [0,360[
    # because we have len(diag.radios)  in [0,720[

    detailed_merit_list = []  # one for each theta theta+360 pair
    detailed_weight_list = []  # one for each theta theta+360 pair

    for i in range(n_angles_2pi):
        # if we have something for both items of the pair, proceed
        if (not np.isnan(diag.zpix_transl[i])) and (not np.isnan(diag.zpix_transl[i + n_angles_2pi])):
            # because we have theta and theta + 360

            zpix_mm = diag.pixel_size_mm
            add, add_w = merit(
                diag.radios[i], diag.radios[i + n_angles_2pi], diag.weights[i], diag.weights[i + n_angles_2pi], shift
            )

            detailed_merit_list.append(add)
            detailed_weight_list.append(add_w)

            observed_oneturn_total_shift_zpix_list.append(diag.zpix_transl[i + n_angles_2pi] - diag.zpix_transl[i])

    return detailed_merit_list, detailed_weight_list, observed_oneturn_total_shift_zpix_list, zpix_mm


def merit(ima_a, ima_b, w_a, w_b, s):
    """A definition of the merit which accounts also for the data weight.
    calculates the merit for a given shift s.
    Comparison between a and b
    Considering signal ima and weight w
    """
    if s == 0:
        # return - abs(  (ima_a - ima_b)  * w_a * w_b      ).astype("d").mean(), (w_a * w_b).astype("d").mean()

        return (ima_a * ima_b * w_a * w_b).astype("d").sum(), (w_a * w_b).astype("d").sum()

    elif s > 0:
        # Keep the comment lines in case one wish to test L1
        # pi = abs(ima_b[s:] - ima_a[:-s])
        # pw = w_b[s:] * w_a[:-s]
        # return -  ( pi * pw      ).astype("d").mean(), (pw).astype("d").mean()

        pi = ima_b[s:] * ima_a[:-s]
        pw = w_b[s:] * w_a[:-s]
        return (pi * pw).astype("d").sum(), pw.astype("d").sum()
    else:
        # Keep the comment lines in case one wish to test L1
        # pi = abs(ima_a[-s:] - ima_b[:s])
        # pw = w_a[-s:] * w_b[:s]
        # return - ( pi * pw ).astype("d").mean(), pw.astype("d").mean()

        pi = ima_a[-s:] * ima_b[:s]
        pw = w_a[-s:] * w_b[:s]
        return (pi * pw).astype("d").sum(), pw.astype("d").sum()


def build_total_merit_list(diag, oversample_factor, args):
    # calculats the merit at all the tested extra adjustment shifts.

    transform_images(diag, [oversample_factor, 1])
    h_ima = diag.radios[0].shape[0]
    # search_radius_v = min(oversample_factor * args.search_radius_v, h_ima - 1)
    search_radius_v = oversample_factor * args.search_radius_v

    shift_s = []
    for_all_shifts_detailed_merit_lists = []
    for_all_shifts_detailed_weight_lists = []

    observed_oneturn_total_shift_zpix_list, zpix_mm = None, None

    for shift in range(-search_radius_v, search_radius_v + 1):
        (
            detailed_merit_list,
            detailed_weight_list,
            found_observed_oneturn_total_shift_zpix_list,
            found_zpix_mm,
        ) = detailed_merit(diag, shift)

        if found_zpix_mm is not None:
            # the following two lines do not depend on the shift.
            # The shift is what we do prior to a comparison f images
            # while the two items below are a properties of the scan
            # in particular they depend on z_translation and angles from bliss
            zpix_mm = found_zpix_mm
            observed_oneturn_total_shift_zpix_list = found_observed_oneturn_total_shift_zpix_list

            # The merit and weight are the result of comparison, they depend on the shift
            for_all_shifts_detailed_merit_lists.append(detailed_merit_list)
            for_all_shifts_detailed_weight_lists.append(detailed_weight_list)
            shift_s.append(
                shift / oversample_factor
            )  # shift_s is stored in original pixel units. Images were oversampled

        else:
            # here there is nothing to append, not correspondance was found
            pass

    # now transposition: we want for each pair  theta, theta+360 a list which contains meritvalues for each adjustment shift
    # For each pair there is a list which runs over the shifts
    # Same thing for the weights
    for_all_pairs_detailed_merit_lists = zip(*for_all_shifts_detailed_merit_lists)
    for_all_pairs_detailed_weight_lists = zip(*for_all_shifts_detailed_weight_lists)

    return (
        for_all_pairs_detailed_merit_lists,
        for_all_pairs_detailed_weight_lists,
        observed_oneturn_total_shift_zpix_list,
        zpix_mm,
    )


def main(user_args=None):
    """Analyse the diagnostics and correct the pixel size"""

    if user_args is None:
        user_args = sys.argv[1:]

    args = DictToObj(
        parse_params_values(
            DiagToPixConfig,
            parser_description=main.__doc__,
            program_version="nabu " + version,
            user_args=user_args,
        )
    )

    oversample_factor = 4
    if args.nexus_source is None:
        args.nexus_source = args.nexus_target

    ## Read all the available diagnostics.
    ## Every key correspond to a chunk of the helical pipeline
    diag_url = os.path.join("/", args.entry_name, "reconstruction/configuration/diagnostics")
    diag_keys = []
    with h5py.File(args.diag_file, "r") as f:
        diag_keys = list(f[diag_url].keys())
        diag_keys = [diag_keys[i] for i in np.argsort(list(map(int, diag_keys)))]

    # The diag_keys are 0,1,2 ... corresponding to all the z-windows ( chunck) for which we have collected contributions at different angles
    # which are nothing else but pieces of ready to use preprocessed radio , and this for different angles.
    # Pairs should coincide where they both have signal ( part of them can be dark if the detector is out of view above or below)
    # For each key there is a sequence of radio, the corresponding sequence of weights map, the corresponding z translation, and angles

    zpix_mm = None
    observed_oneturn_total_shift_zpix = None

    argument_list = [
        (DictToObj(h5todict(args.diag_file, os.path.join(diag_url, my_key))), oversample_factor, args)
        for my_key in diag_keys
    ]

    ncpus = get_available_threads()
    with Pool(processes=ncpus) as pool:
        all_res_plus_infos = pool.starmap(build_total_merit_list, argument_list)

    observed_oneturn_total_shift_zpix, zpix_mm = None, None

    # needs to flatten the result of pool.map
    for_all_pairs_detailed_merit_lists = []
    for_all_pairs_detailed_weight_lists = []
    observed_oneturn_total_shift_zpix_list = []
    zpix_mm = None

    for (
        tmp_merit_lists,
        tmp_weight_lists,
        tmp_observed_oneturn_total_shift_zpix_list,
        tmp_zpix_mm,
    ) in all_res_plus_infos:
        if tmp_zpix_mm is not None:
            # then each item of the composed list will be for a given pairs theta, theta+360
            # and each such item is a list where each item is for a given probed shift
            for_all_pairs_detailed_merit_lists.extend(tmp_merit_lists)
            for_all_pairs_detailed_weight_lists.extend(tmp_weight_lists)
            observed_oneturn_total_shift_zpix_list.extend(tmp_observed_oneturn_total_shift_zpix_list)

            zpix_mm = tmp_zpix_mm

    if zpix_mm is None:
        message = "No overlapping was found"
        raise RuntimeError(message)

    if len(for_all_pairs_detailed_merit_lists) == 0:
        message = "No diag was found"
        raise RuntimeError(message)

    # Now an important search step:
    # We find for which pair of theta theta+360 the observed translation has the bigger absolute value.
    # Then the search for the optimum is performed for the readjustment shift in the
    # range (-search_radius_v, search_radius_v + 1)
    # considered as readjustmnet for the foud ideal pair which has exactly a translation equal to this  maximal absolute observed translation
    # For all the others the readjustment is multiplied by the pro-rata factor
    # given by their smaller z-translation over the maximal one
    max_absolute_shift = abs(np.array(observed_oneturn_total_shift_zpix_list)).max()

    # gong to search for the best pixel size
    max_merit = None
    best_shift = None
    search_radius_v = oversample_factor * args.search_radius_v

    probed_shift_list = list(range(-search_radius_v, search_radius_v + 1))

    for shift in range(-search_radius_v, search_radius_v + 1):
        total_sum = 0
        total_weight = 0

        for merit_list, weight_list, one_turn_shift in zip(
            for_all_pairs_detailed_merit_lists,
            for_all_pairs_detailed_weight_lists,
            observed_oneturn_total_shift_zpix_list,
        ):
            # sanity check
            assert len(merit_list) == len(probed_shift_list), " this should not happen"
            assert len(weight_list) == len(probed_shift_list), " this should not happen"

            # pro_rata shift
            my_shift = shift * (one_turn_shift / max_absolute_shift)

            # doing interpolation with search sorted
            i1 = np.searchsorted(probed_shift_list, my_shift)

            if i1 > 0 and i1 < len(probed_shift_list):
                i0 = i1 - 1
                fract = (-my_shift + probed_shift_list[i1]) / (probed_shift_list[i1] - probed_shift_list[i0])

                total_sum += fract * merit_list[i0] + (1 - fract) * merit_list[i1]
                total_weight += fract * weight_list[i0] + (1 - fract) * weight_list[i1]

        if total_weight == 0:
            # this avoid 0/0 = nan
            total_weight = 1

        m = total_sum / total_weight

        if (max_merit is None) or ((not np.isnan(m)) and m > max_merit):
            max_merit = m
            best_shift = shift / oversample_factor

    print(" Best shift at ", best_shift)
    print(
        f" Over one turn the reference  shift was {max_absolute_shift}  pixels. But a residual shift of {best_shift} remains   "
    )
    # the formula below is already purged from the ovrsamplig factor. We did this when we recorded best_shift and the z shift
    # is registered when lloking at the z_translation and one does not fiddle aroud with  the oversamplig at that moment
    zpix_mm = zpix_mm * (max_absolute_shift) / (max_absolute_shift - best_shift)

    print(f"Corrected zpix_mm = {zpix_mm}")

    if args.nexus_target is not None:
        nx_tomo = NXtomo().load(args.nexus_source, args.entry_name)
        nx_tomo.instrument.detector.x_pixel_size = zpix_mm * 1.0e-3  # pixel size must be provided in SI  (meters)
        nx_tomo.instrument.detector.y_pixel_size = zpix_mm * 1.0e-3  # pixel size must be provided in SI  (meters)
        nx_tomo.save(file_path=args.nexus_target, data_path=args.entry_name, overwrite=True)

    return 0



import pytest
import numpy as np
from nabu.testutils import get_data, __do_long_tests__

from nabu.cuda.utils import __has_pycuda__
from nabu.reconstruction.mlem import MLEMReconstructor, __have_corrct__


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    datafile = get_data("sl_mlem.npz")
    cls.data = datafile["data"]
    cls.angles_rad = datafile["angles_rad"]
    cls.random_u_shifts = datafile["random_u_shifts"]
    cls.ref_rec_noshifts = datafile["ref_rec_noshifts"]
    cls.ref_rec_shiftsu = datafile["ref_rec_shiftsu"]
    cls.ref_rec_u_rand = datafile["ref_rec_u_rand"]
    cls.ref_rec_shiftsv = datafile["ref_rec_shiftsv"]
    # cls.ref_rec_v_rand = datafile["ref_rec_v_rand"]
    cls.tol = 2e-4


@pytest.mark.skipif(not (__has_pycuda__ and __have_corrct__), reason="Need pycuda and corrct for this test")
@pytest.mark.usefixtures("bootstrap")
class TestMLEM:
    """These tests test the general MLEM reconstruction algorithm
    and the behavior of the reconstruction with respect to horizontal shifts.
    Only horizontal shifts are tested here because vertical shifts are handled outside
    the reconstruction object, but in the embedding reconstruction pipeline. See FullFieldReconstructor"""

    def _create_MLEM_reconstructor(self, shifts_uv=None):
        return MLEMReconstructor(
            self.data.shape, -self.angles_rad, shifts_uv, cor=0.0, n_iterations=10  # mind the sign
        )

    def test_simple_mlem_recons(self):
        R = self._create_MLEM_reconstructor()
        rec = R.reconstruct(self.data)
        delta = np.abs(rec[:, ::-1, :] - self.ref_rec_noshifts)
        assert np.max(delta) < self.tol

    def test_mlem_recons_with_u_shifts(self):
        shifts = np.zeros((len(self.angles_rad), 2))
        shifts[:, 0] = -5
        R = self._create_MLEM_reconstructor(shifts)
        rec = R.reconstruct(self.data)
        delta = np.abs(rec[:, ::-1] - self.ref_rec_shiftsu)
        assert np.max(delta) < self.tol

    def test_mlem_recons_with_random_u_shifts(self):
        R = self._create_MLEM_reconstructor(self.random_u_shifts)
        rec = R.reconstruct(self.data)
        delta = np.abs(rec[:, ::-1] - self.ref_rec_u_rand)
        assert np.max(delta) < self.tol

    def test_mlem_recons_with_constant_v_shifts(self):
        from nabu.preproc.shift import VerticalShift

        shifts = np.zeros((len(self.angles_rad), 2))
        shifts[:, 1] = -20

        nv, n_angles, nu = self.data.shape
        radios_movements = VerticalShift(
            (n_angles, nv, nu), -shifts[:, 1]
        )  # Minus sign here mimics what is  done in the pipeline.
        tmp_in = np.swapaxes(self.data, 0, 1).copy()
        tmp_out = np.zeros_like(tmp_in)
        radios_movements.apply_vertical_shifts(tmp_in, list(range(n_angles)), output=tmp_out)
        data = np.swapaxes(tmp_out, 0, 1).copy()

        R = self._create_MLEM_reconstructor(shifts)
        rec = R.reconstruct(data)

        axslice = 120
        trslice = 84
        axslice1 = self.ref_rec_shiftsv[axslice]
        axslice2 = rec[axslice, ::-1]
        trslice1 = self.ref_rec_shiftsv[trslice]
        trslice2 = rec[trslice, ::-1]
        # delta = np.abs(rec[:, ::-1] - self.ref_rec_shiftsv)
        delta_ax = np.abs(axslice1 - axslice2)
        delta_tr = np.abs(trslice1 - trslice2)
        assert max(np.max(delta_ax), np.max(delta_tr)) < self.tol

    @pytest.mark.skip(reason="No valid reference reconstruction for this test.")
    def test_mlem_recons_with_random_v_shifts(self):
        """NOT YET IMPLEMENTED.
        This is a temporary version due to unpexcted behavior of CorrCT/Astra to
        compute a reference implementation. See [question on Astra's github](https://github.com/astra-toolbox/astra-toolbox/discussions/520).
        """



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../"))


# -- Project information -----------------------------------------------------

project = "Nabu"
copyright = "2019-2024, ESRF"
author = "Pierre Paleo"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "nbsphinx",
    #    'sphinx.ext.autosummary',
    #    'sphinx.ext.doctest',
    #    'sphinx.ext.inheritance_diagram',
]
# myst_commonmark_only = True
# for myst
suppress_warnings = [
    "myst.header",  # non-consecutive headers levels
    "autosectionlabel.*",  # duplicate section names
]
myst_heading_anchors = 3
myst_enable_extensions = [
    #    "amsmath",
    # #   "colon_fence",
    # #   "deflist",
    "dollarmath",
    #    "html_admonition",
    #    "html_image",
    #    "linkify",
    # #   "replacements",
    # #   "smartquotes",
    # #   "substitution"
]
#
# autosummary_generate = True
autodoc_member_order = "bysource"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

# from corlab_theme import get_theme_dir
# html_theme = 'corlab_theme'

from cloud_sptheme import get_theme_dir

html_theme = "cloud"

html_theme_path = [get_theme_dir()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "theme_overrides.css",
]

html_theme_options = {
    #    'navigation_depth': -1,
    "max_width": "75%",
    "minimal_width": "720px",
}


# For mathjax
mathjax_path = "javascript/MathJax-3.0.5/es5/tex-mml-chtml.js"


"""
# For recommonmark
from recommonmark.transform import AutoStructify
github_doc_root = 'https://github.com/rtfd/recommonmark/tree/master/doc/'
def setup(app):
    app.add_config_value('recommonmark_config', {
            'url_resolver': lambda url: github_doc_root + url,
            'auto_toc_tree_section': 'Contents',
            'enable_math': True,
            'enable_inline_math': True,
            }, True)
    app.add_transform(AutoStructify)
"""

# Document __init__
autoclass_content = "both"

from nabu import __version__

version = __version__
release = version

master_doc = "index"

#
# nbsphinx
#

nbsphinx_allow_errors = True



import os

import h5py
import numpy
import pytest
from tqdm import tqdm
from silx.image.phantomgenerator import PhantomGenerator
from tomoscan.esrf.volume import EDFVolume, HDF5Volume
from tomoscan.esrf.volume.tiffvolume import TIFFVolume, has_tifffile
from tomoscan.factory import Factory as TomoscanFactory
from tomoscan.utils.volume import concatenate as concatenate_volumes

from nabu.stitching.alignment import AlignmentAxis1, AlignmentAxis2
from nabu.stitching.config import NormalizationBySample, PostProcessedZStitchingConfiguration
from nabu.stitching.overlap import OverlapStitchingStrategy
from nabu.stitching.utils import ShiftAlgorithm
from nabu.stitching.z_stitching import PostProcessZStitcher, PostProcessZStitcherNoDD


strategies_to_test_weights = (
    OverlapStitchingStrategy.CLOSEST,
    OverlapStitchingStrategy.COSINUS_WEIGHTS,
    OverlapStitchingStrategy.LINEAR_WEIGHTS,
    OverlapStitchingStrategy.MEAN,
)


def build_raw_volume():
    """util to create some raw volume"""
    raw_volume = numpy.stack(
        [
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 256.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 128.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 32.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 16.0,
        ]
    )
    assert raw_volume.shape == (4, 120, 120)
    raw_volume = numpy.rollaxis(raw_volume, axis=1, start=0)
    assert raw_volume.shape == (120, 4, 120)
    return raw_volume


_VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING = [HDF5Volume, EDFVolume]
# avoid testing glymur because doesn't handle float
# if has_minimal_openjpeg:
#     _VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING.append(JP2KVolume)
if has_tifffile:
    _VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING.append(TIFFVolume)


def build_volumes(output_dir: str, volume_class):
    # create some random data.
    raw_volume = build_raw_volume()

    # create a simple case where the volume have 10 voxel of overlap and a height (z) of 30 Voxels, 40 and 30 Voxels
    vol_1_constructor_params = {
        "data": raw_volume[0:30, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_2_constructor_params = {
        "data": raw_volume[20:80, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_3_constructor_params = {
        "data": raw_volume[60:, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    volumes = []
    axis_0_positions = []
    for i_vol, vol_params in enumerate([vol_1_constructor_params, vol_2_constructor_params, vol_3_constructor_params]):
        if volume_class == HDF5Volume:
            vol_params.update(
                {
                    "file_path": os.path.join(output_dir, f"raw_volume_{i_vol}.hdf5"),
                    "data_path": "volume",
                }
            )
        else:
            vol_params.update(
                {
                    "folder": os.path.join(output_dir, f"raw_volume_{i_vol}"),
                }
            )
        axis_0_positions.append(vol_params["metadata"]["processing_options"]["reconstruction"]["position"][0])

        volume = volume_class(**vol_params)
        volume.save()
        volumes.append(volume)
    return volumes, axis_0_positions, raw_volume


@pytest.mark.parametrize("progress", (None, "with_tqdm"))
@pytest.mark.parametrize("volume_class", (_VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING))
def test_PostProcessZStitcher(
    tmp_path,
    volume_class,
    progress,
):
    """
    test PreProcessZStitcher class and insure a full stitching can be done automatically.

    :param bool clear_input_volumes_data: if True save the volume then clear volume.data (used to check internal management of loading volumes - used to check behavior with HDF5)
    :param volume_class: class to be used (same class for input and output for now)
    :param axis_0_pos: position of the different TomoObj along axis 0 (Also know as z axis)
    """
    if progress == "with_tqdm":
        progress = tqdm(total=100)

    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    volumes, axis_0_positions, raw_volume = build_volumes(output_dir=raw_data_dir, volume_class=volume_class)
    volume_1, volume_2, volume_3 = volumes

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=axis_0_positions,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
    )

    stitcher = PostProcessZStitcher(z_stich_config, progress=progress)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.data = None
    output_volume.metadata = None
    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert raw_volume.shape == output_volume.data.shape

    numpy.testing.assert_array_almost_equal(raw_volume, output_volume.data)

    metadata = output_volume.metadata
    assert "about" in metadata
    assert "configuration" in metadata
    assert output_volume.position[0] == -60.0
    assert output_volume.pixel_size == (1.0, 1.0, 1.0)


slices_to_test_post = (
    {
        "slices": (None,),
        "complete": True,
    },
    {
        "slices": (("first",), ("middle",), ("last",)),
        "complete": False,
    },
    {
        "slices": ((0, 1, 2), slice(3, -1, 1)),
        "complete": True,
    },
)


@pytest.mark.parametrize("flip_ud", (True, False))
@pytest.mark.parametrize("configuration_dist", slices_to_test_post)
def test_DistributePostProcessZStitcher(tmp_path, configuration_dist, flip_ud):
    # create some random data.
    slices = configuration_dist["slices"]
    complete = configuration_dist["complete"]

    raw_volume = numpy.ones((80, 40, 120), dtype=numpy.float16)
    raw_volume[:, 0, :] = (
        PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float16)[30:110, :] * 80 * 40 * 120
    )
    raw_volume[:, 8, :] = (
        PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float16)[30:110, :] * 80 * 40 * 120 + 2
    )
    raw_volume[12] = 1.0
    raw_volume[:, 23] = 1.2
    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    def flip_input_data(data):
        if flip_ud:
            data = numpy.flipud(data)
        return data

    volume_1 = HDF5Volume(
        file_path=os.path.join(raw_data_dir, "volume_1.hdf5"),
        data_path="volume",
        data=flip_input_data(raw_volume[-60:, :, :]),
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-30.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    )
    volume_1.save()

    volume_2 = HDF5Volume(
        file_path=os.path.join(raw_data_dir, "volume_2.hdf5"),
        data_path="volume",
        data=flip_input_data(raw_volume[:60, :, :]),
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    )
    volume_2.save()

    reconstructed_sub_volumes = []
    for i_slice, s in enumerate(slices):
        output_volume = HDF5Volume(
            file_path=os.path.join(output_dir, f"stitched_subvolume_{i_slice}.hdf5"),
            data_path="stitched_volume",
        )
        volumes = (volume_2, volume_1)
        z_stich_config = PostProcessedZStitchingConfiguration(
            stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
            axis_0_pos_px=tuple(
                volume.metadata["processing_options"]["reconstruction"]["position"][0] for volume in volumes
            ),
            axis_0_pos_mm=None,
            axis_0_params={},
            axis_1_pos_px=(0, 0),
            axis_1_pos_mm=None,
            axis_1_params={},
            axis_2_pos_px=None,
            axis_2_pos_mm=None,
            axis_2_params={},
            overwrite_results=True,
            input_volumes=volumes,
            output_volume=output_volume,
            slices=s,
            slurm_config=None,
            slice_for_cross_correlation="middle",
            voxel_size=None,
            flip_ud=flip_ud,
        )

        stitcher = PostProcessZStitcher(z_stich_config)
        vol_id = stitcher.stitch()
        reconstructed_sub_volumes.append(TomoscanFactory.create_tomo_object_from_identifier(identifier=vol_id))

    final_vol = HDF5Volume(
        file_path=os.path.join(output_dir, "final_volume"),
        data_path="volume",
    )
    if complete:
        concatenate_volumes(output_volume=final_vol, volumes=tuple(reconstructed_sub_volumes), axis=1)
        final_vol.load_data(store=True)
        numpy.testing.assert_almost_equal(
            raw_volume,
            final_vol.data,
        )


@pytest.mark.parametrize("alignment_axis_2", ("left", "right", "center"))
def test_vol_z_stitching_with_alignment_axis_2(tmp_path, alignment_axis_2):
    """
    test z volume stitching with different width (and so that requires image alignment over axis 2)
    """
    # create some random data.
    raw_volume = build_raw_volume()
    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    # create a simple case where the volume have 10 voxel of overlap and a height (z) of 30 Voxels, 40 and 30 Voxels
    vol_1_constructor_params = {
        "data": raw_volume[0:30, :, 4:-4],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_2_constructor_params = {
        "data": raw_volume[20:80, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_3_constructor_params = {
        "data": raw_volume[60:, :, 10:-10],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    raw_volumes = []
    axis_0_positions = []
    for i_vol, vol_params in enumerate([vol_1_constructor_params, vol_2_constructor_params, vol_3_constructor_params]):
        vol_params.update(
            {
                "file_path": os.path.join(raw_data_dir, f"raw_volume_{i_vol}.hdf5"),
                "data_path": "volume",
            }
        )
        axis_0_positions.append(vol_params["metadata"]["processing_options"]["reconstruction"]["position"][0])
        volume = HDF5Volume(**vol_params)
        volume.save()
        raw_volumes.append(volume)

    volume_1, volume_2, volume_3 = raw_volumes

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=axis_0_positions,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
        alignment_axis_2=AlignmentAxis2.from_value(alignment_axis_2),
    )

    stitcher = PostProcessZStitcher(z_stich_config, progress=None)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert output_volume.data.shape == (120, 4, 120)

    if alignment_axis_2 == "center":
        numpy.testing.assert_array_almost_equal(raw_volume[:, :, 10:-10], output_volume.data[:, :, 10:-10])
    elif alignment_axis_2 == "left":
        numpy.testing.assert_array_almost_equal(raw_volume[:, :, :-20], output_volume.data[:, :, :-20])
    elif alignment_axis_2 == "right":
        numpy.testing.assert_array_almost_equal(raw_volume[:, :, 20:], output_volume.data[:, :, 20:])


@pytest.mark.parametrize("alignment_axis_1", ("front", "center", "back"))
def test_vol_z_stitching_with_alignment_axis_1(tmp_path, alignment_axis_1):
    """
    test z volume stitching with different number of frames (and so that requires image alignment over axis 0)
    """
    # create some random data.
    raw_volume = build_raw_volume()

    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    # create a simple case where the volume have 10 voxel of overlap and a height (z) of 30 Voxels, 40 and 30 Voxels
    vol_1_constructor_params = {
        "data": raw_volume[
            0:30,
            1:3,
        ],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_2_constructor_params = {
        "data": raw_volume[20:80, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_3_constructor_params = {
        "data": raw_volume[
            60:,
            1:3,
        ],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    raw_volumes = []
    axis_0_positions = []
    for i_vol, vol_params in enumerate([vol_1_constructor_params, vol_2_constructor_params, vol_3_constructor_params]):
        vol_params.update(
            {
                "file_path": os.path.join(raw_data_dir, f"raw_volume_{i_vol}.hdf5"),
                "data_path": "volume",
            }
        )
        axis_0_positions.append(vol_params["metadata"]["processing_options"]["reconstruction"]["position"][0])
        volume = HDF5Volume(**vol_params)
        volume.save()
        raw_volumes.append(volume)

    volume_1, volume_2, volume_3 = raw_volumes

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=axis_0_positions,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
        alignment_axis_1=AlignmentAxis1.from_value(alignment_axis_1),
    )

    stitcher = PostProcessZStitcher(z_stich_config, progress=None)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert output_volume.data.shape == (120, 4, 120)

    if alignment_axis_1 == "middle":
        numpy.testing.assert_array_almost_equal(raw_volume[:, 10:-10, :], output_volume.data[:, 10:-10, :])
    elif alignment_axis_1 == "front":
        numpy.testing.assert_array_almost_equal(raw_volume[:, :-20, :], output_volume.data[:, :-20, :])
    elif alignment_axis_1 == "middle":
        numpy.testing.assert_array_almost_equal(raw_volume[:, 20:, :], output_volume.data[:, 20:, :])


def test_normalization_by_sample(tmp_path):
    """
    simple test of a volume stitching.
    Raw volumes have 'extra' values (+2, +5, +9) that must be removed at the end thanks to the normalization
    """
    raw_volume = build_raw_volume()
    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    # create a simple case where the volume have 10 voxel of overlap and a height (z) of 30 Voxels, 40 and 30 Voxels
    vol_1_constructor_params = {
        "data": raw_volume[0:30, :, :] + 3,
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_2_constructor_params = {
        "data": raw_volume[20:80, :, :] + 5,
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_3_constructor_params = {
        "data": raw_volume[60:, :, :] + 12,
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    raw_volumes = []
    axis_0_positions = []
    for i_vol, vol_params in enumerate([vol_1_constructor_params, vol_2_constructor_params, vol_3_constructor_params]):
        vol_params.update(
            {
                "file_path": os.path.join(raw_data_dir, f"raw_volume_{i_vol}.hdf5"),
                "data_path": "volume",
            }
        )

        axis_0_positions.append(vol_params["metadata"]["processing_options"]["reconstruction"]["position"][0])

        volume = HDF5Volume(**vol_params)
        volume.save()
        raw_volumes.append(volume)

    volume_1, volume_2, volume_3 = raw_volumes

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    normalization_by_sample = NormalizationBySample()
    normalization_by_sample.set_is_active(True)
    normalization_by_sample.width = 1
    normalization_by_sample.margin = 0
    normalization_by_sample.side = "left"
    normalization_by_sample.method = "median"

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.CLOSEST,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=axis_0_positions,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
        normalization_by_sample=normalization_by_sample,
    )

    stitcher = PostProcessZStitcher(z_stich_config, progress=None)
    output_identifier = stitcher.stitch()

    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.data = None
    output_volume.metadata = None
    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert raw_volume.shape == output_volume.data.shape

    numpy.testing.assert_array_almost_equal(raw_volume, output_volume.data)

    metadata = output_volume.metadata
    assert "configuration" in metadata
    assert "about" in metadata
    assert metadata["about"]["program"] == "nabu-stitching"
    assert output_volume.position[0] == -60.0
    assert output_volume.pixel_size == (1.0, 1.0, 1.0)


@pytest.mark.parametrize("data_duplication", (True, False))
def test_data_duplication(tmp_path, data_duplication):
    """
    Test that the post-processing stitching can be done without duplicating data.
    And also making sure avoid data duplication can handle frame flips
    """
    raw_volume = build_raw_volume()

    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    volume_1 = HDF5Volume(
        data=raw_volume[0:30],
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
        file_path=os.path.join(raw_data_dir, f"raw_volume_1.hdf5"),
        data_path="volume",
    )

    volume_2 = HDF5Volume(
        data=raw_volume[20:80],
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
        file_path=os.path.join(raw_data_dir, f"raw_volume_2.hdf5"),
        data_path="volume",
    )

    volume_3 = HDF5Volume(
        data=raw_volume[60:],
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
        file_path=os.path.join(raw_data_dir, f"raw_volume_3.hdf5"),
        data_path="volume",
    )

    for volume in (volume_1, volume_2, volume_3):
        volume.save()
        volume.clear_cache()

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.CLOSEST,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=None,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
        duplicate_data=data_duplication,
    )

    if data_duplication:
        stitcher = PostProcessZStitcher(z_stich_config, progress=None)
    else:
        stitcher = PostProcessZStitcherNoDD(z_stich_config, progress=None)
    output_identifier = stitcher.stitch()

    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.data = None
    output_volume.metadata = None
    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert raw_volume.shape == output_volume.data.shape
    numpy.testing.assert_almost_equal(raw_volume.data, output_volume.data)

    with h5py.File(output_volume.file_path, mode="r") as h5f:
        if data_duplication:
            assert f"{output_volume.data_path}/stitching_regions" not in h5f
            assert not h5f[f"{output_volume.data_path}/results/data"].is_virtual
        else:
            assert f"{output_volume.data_path}/stitching_regions" in h5f
            assert h5f[f"{output_volume.data_path}/results/data"].is_virtual

    if not data_duplication:
        # make sure an error is raised if we try to ask for no data duplication and if we get some flips
        z_stich_config.flip_ud = (False, True, False)
        with pytest.raises(ValueError):
            stitcher = PostProcessZStitcherNoDD(z_stich_config, progress=None)
            stitcher.stitch()
